import pandas as pd
import modin.pandas as pandas
import numpy as np
import matplotlib.pyplot as plt
import datetime as dt
import json
import functools
from ast import literal_eval
import seaborn as sns

import load_data, data_support, graph_support

# Set to false to recreate numbers. Warning - long runtime. Does not recreate base csv
# Recreates graphs anyway
use_cache = True

#Use to skip loading after 1000 entries, will use different cache than non-testrun
testrun = True

onlyUser = True

def userOrSender():
    global onlyUser
    if onlyUser:
        return ['user']
    else:
        return ['user', 'sender']

YEAR = 365

def cache_it(cache_name, is_df=True):
    def real_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if is_cached(cache_name):
                if is_df:
                    result = load_from_cache(cache_name)
                else:
                    with open(load_data.get_cache_path(cache_name, testrun), 'r') as fp:
                        result = fp.read()
            else:
                result = function(*args, **kwargs)
                if is_df:
                    store_to_cache(cache_name, result)
                else:
                    with open(load_data.get_cache_path(cache_name, testrun), 'w') as fp:
                        fp.write(str(result))
            return result
        return wrapper
    return real_decorator

def store_to_cache(name, data):
    global testrun
    return load_data.store_to_cache(name, data, testrun)

def is_cached(name):
    global use_cache
    global testrun
    return load_data.is_cached(name, testrun, use_cache)

def load_from_cache(name):
    global testrun
    return load_data.load_from_cache(name, testrun)

@cache_it("oliver_smime_university_communication.csv", is_df=True)
def smime_communication(df_mails = None):
    """
        Input: df of all mails and df of all smime users at the university
        Output a df with the smime communication between users
        
        user 1 (from) | email user 1 | uses smime (bool) | (to) user 2 | count mails u1 to u2 | count mails u2 to u1 | count signed mails u1 to u2 | first sent cert from u1 to u2 (date) | unsigned mails since first cert u1 to u2 | encrypted mails from u1 to u2 | first encrypted mail from u1 to u2 | unencrypted mails since first enc mail u1 to u2
        Remove not mails form the receiver.
    """
    global testrun
    if df_mails == None:
        df_mails = load_data.load_all_mails(testrun, testrows = 5000000)
        df_mails = load_data.drop_error_mails(df_mails)
    df_mails = data_support.add_level_info_fix_crypto(df_mails)

    df = df_mails[['user', 'sender', 'receiver_list', 'cc_list', 'bcc_list', 'folder', 'header_date', 'cert_data', 'smime_signature', 'smime_encryption', 'pgp_encryption_attachment', 'pgp_signature_attachment', 'pgp_key_data_attachment', 'smime_sig_fixed','smime_level', 'is_sent']]
    
    df = data_support.get_sent_mails(df)
    # only mails that include a university email address
    df = df.loc[df['receiver_list'].str.contains('.h') | df['cc_list'].str.contains('.h') | df['bcc_list'].str.contains('.h')]
    
    print("> start to split receiver lists.")
    df['receiver_list'] = df.receiver_list.apply(literal_eval)
    df = df.explode('receiver_list')
    df['cc_list'] = df.cc_list.apply(literal_eval)
    df = df.explode('cc_list')
    df['bcc_list'] = df.bcc_list.apply(literal_eval)
    df = df.explode('bcc_list')

    df = df.melt(id_vars = ['user', 'sender', 'header_date', 'cert_data', 'smime_signature', 'smime_encryption', 'smime_sig_fixed','smime_level', 'is_sent'], value_vars= ['receiver_list', 'cc_list', 'bcc_list'], var_name = 'receiver_type', value_name = 'receiver')
    df = df[df['receiver'].notna()].reset_index()
    df = df.loc[df['receiver'].str.contains('.h')]
    df['header_date'] = pandas.to_datetime(df['header_date'], format="%Y-%m-%d", errors='coerce')

    user_receiver = df[['user', 'sender', 'receiver']].drop_duplicates()
    user_receiver = user_receiver.dropna()
    user_receiver = create_communication_df(df, user_receiver)
       
    # Add date when the first signed mail is sent from the receiver to the sender (from receiver as sender)
    user_receiver = user_receiver.drop_duplicates()
    user_receiver = user_receiver.merge(user_receiver, how='left', left_on=['receiver','sender'], right_on=['sender', 'receiver'], suffixes=('', '_from_receiver'))
    # count all mails from sender to receiver after date first_mail_receiver
    print("> count all mails between a sender to receiver after the first mail from the receiver.")
    df = df.merge(user_receiver[['user', 'sender', 'receiver', 'first_mail_from_receiver', 'signed_first_mail_from_receiver', 'enc_first_mail_from_receiver']], how='left', on=['user','sender','receiver'])
    # First received mail ever
    user_receiver = add_mails_after(df, user_receiver, 'first_mail_from_receiver')
    # First signed received mail
    user_receiver = add_mails_after(df, user_receiver,  'signed_first_mail_from_receiver')
    # First enc received mail
    user_receiver = add_mails_after(df, user_receiver, 'enc_first_mail_from_receiver')
    print("> before remove duplicates: ", user_receiver.shape)
    user_receiver = user_receiver.drop_duplicates(subset=['user','sender', 'receiver'])
    return user_receiver
    

def create_communication_df(df, user_receiver):
    print("> count all mails from a sender to receiver.")
    mails_between_user_sender = df.groupby(userOrSender() + ['receiver']).agg(mails = ('header_date', 'count'), first_mail = ('header_date', 'min'), last_mail = ('header_date', 'max')).reset_index()
    user_receiver = user_receiver.merge(mails_between_user_sender, how = 'left', on = userOrSender() + ['receiver'])
    
    df2 = df.loc[df['smime_sig_fixed'] == True]
    mails_between_user_sender = df2.groupby(userOrSender() + ['receiver']).agg(signed_mails = ('header_date', 'count'), signed_first_mail = ('header_date', 'min'), signed_last_mail = ('header_date', 'max'))
    if isinstance(mails_between_user_sender, pandas.DataFrame):
        user_receiver = user_receiver.merge(mails_between_user_sender, how = 'left', on = userOrSender() + ['receiver'])
        
    df2 = df.loc[df['smime_encryption'] == 'SMIMEEnc.enc']
    mails_between_user_sender = df2.groupby(userOrSender() + ['receiver']).agg(enc_mails = ('header_date', 'count'), enc_first_mail = ('header_date', 'min'), enc_last_mail = ('header_date', 'max'))
    if isinstance(mails_between_user_sender, pandas.DataFrame):
        user_receiver = user_receiver.merge(mails_between_user_sender, how = 'left', on = userOrSender() + ['receiver'])
        
    return user_receiver
    
def add_mails_after(df, user_receiver, first_mail_of):
    counterName = 'sent_mails_after_'
    # Count mails after first response
    df_two_way_com = df[df[first_mail_of].notna()]
    user_receiver = help_add_mails_after(df_two_way_com, user_receiver, first_mail_of, counterName)
    # Count signed mails after first response
    df2 = df_two_way_com.loc[df_two_way_com['smime_sig_fixed'] == True]
    counterName = 'sent_signed_mails_after_'
    user_receiver = help_add_mails_after(df2, user_receiver, first_mail_of, counterName)
    # Count enc mails after first response
    df2 = df_two_way_com.loc[df_two_way_com['smime_encryption'] == 'SMIMEEnc.enc']
    counterName = 'sent_enc_mails_after_'
    user_receiver = help_add_mails_after(df2, user_receiver, first_mail_of, counterName)
    
    return user_receiver
    
def help_add_mails_after(df, user_receiver, first_mail_of, counterName):
    mails_between_user_sender = df.groupby(userOrSender() + ['receiver']).apply(lambda g: count_mails_after(g, first_mail_of))
    mails_between_user_sender = mails_between_user_sender.rename(columns={'header_date': counterName+first_mail_of})
    return user_receiver.merge(mails_between_user_sender, how = 'left', on = userOrSender() + ['receiver'])
    
def count_mails_after(g, name):
    x = (g.loc[g['header_date'] >= g[name], ['header_date']])
    return x.count()

def analysis_smime_communication(data = None):
    """
    df of smime communication between sender and receiver. Both are from the university
    'Unnamed: 0', 'user', 'sender', 'receiver',
    About sender to receiver:
    ++++++++++++++++++++++++
    
     'mails', 'first_mail', 'last_mail',
     'signed_mails', 'signed_first_mail', 'signed_last_mail',
     'enc_mails', 'enc_first_mail', 'enc_last_mail',
     'sent_mails_after_first_mail_from_receiver', 'sent_signed_mails_after_first_mail_from_receiver', 'sent_enc_mails_after_first_mail_from_receiver',
     'sent_mails_after_signed_first_mail_from_receiver', 'sent_signed_mails_after_signed_first_mail_from_receiver', 'sent_enc_mails_after_signed_first_mail_from_receiver',
     'sent_mails_after_enc_first_mail_from_receiver', 'sent_signed_mails_after_enc_first_mail_from_receiver', 'sent_enc_mails_after_enc_first_mail_from_receiver'
          
    About receiver to sender:
    +++++++++++++++++++++++++
    'user_from_receiver', 'sender_from_receiver','receiver_from_receiver',  <- just for validation
     'mails_from_receiver', 'first_mail_from_receiver', 'last_mail_from_receiver',
    'signed_mails_from_receiver', 'signed_first_mail_from_receiver', 'signed_last_mail_from_receiver',
    'enc_mails_from_receiver', 'enc_first_mail_from_receiver', 'enc_last_mail_from_receiver',
     
    """
    display = pandas.options.display

    display.max_columns = 10
    display.max_rows = 100
    display.max_colwidth = 30
    display.width = 300
    res = {}
    
    df = smime_communication(data)
    print('Before drop shape: ', df.shape)
    df = df.drop_duplicates()
    df = df[df['receiver'].notna()]
    res['countCommunicationPairs'] = "{:,}".format(df.shape[0])
    res['totalSenders'] = "{:,}".format(df['sender'].nunique())
    res['totalSmimeSenders'] = "{:,}".format(df[df['signed_mails'] > 0]['sender'].nunique())
    res['totalUser'] = "{:,}".format(df['user'].nunique())
    res['totalSmimeUser'] = "{:,}".format(df[df['signed_mails'] > 0]['user'].nunique())
    df.replace(np.nan, 0, inplace=True)

    # sender to receiver sent signed mails -> signed_mails > 0
    # receiver to sendter sent signed mails -> signed_mails_from_receiver > 0
    # Caution double counting!
    smime_pairs = df.loc[(df['signed_mails'] > 0) & (df['signed_mails_from_receiver'] > 0)]
    smime_pairs['SignedMailsRatio'] = smime_pairs['sent_signed_mails_after_signed_first_mail_from_receiver'] / smime_pairs['sent_mails_after_signed_first_mail_from_receiver'] * 100
    smime_pairs['EncMailsRatio'] = smime_pairs['sent_enc_mails_after_signed_first_mail_from_receiver'] / smime_pairs['sent_mails_after_signed_first_mail_from_receiver'] * 100
    
    smime_pairs['SignedMailsRatioEnc'] = smime_pairs['sent_signed_mails_after_enc_first_mail_from_receiver'] / smime_pairs['sent_mails_after_enc_first_mail_from_receiver'] * 100
    smime_pairs['EncMailsRatioEnc'] = smime_pairs['sent_enc_mails_after_enc_first_mail_from_receiver'] / smime_pairs['sent_mails_after_enc_first_mail_from_receiver'] * 100
    
    print('> signed pairs: \n', smime_pairs[['user', 'sender', 'receiver', 'sent_signed_mails_after_signed_first_mail_from_receiver', 'sent_enc_mails_after_signed_first_mail_from_receiver', 'EncMailsRatio', 'SignedMailsRatio']])
    
    # About encrypted communications partners
    enc_pairs = df.loc[(df['enc_mails'] > 0) & (df['enc_mails_from_receiver'] > 0)]
    enc_pairs['SignedMailsRatio'] = enc_pairs['sent_signed_mails_after_signed_first_mail_from_receiver'] / enc_pairs['sent_mails_after_signed_first_mail_from_receiver'] * 100
    enc_pairs['EncMailsRatio'] = enc_pairs['sent_enc_mails_after_signed_first_mail_from_receiver'] / enc_pairs['sent_mails_after_signed_first_mail_from_receiver'] * 100
    enc_pairs['SignedMailsRatioEnc'] = enc_pairs['sent_signed_mails_after_enc_first_mail_from_receiver'] / enc_pairs['sent_mails_after_enc_first_mail_from_receiver'] * 100
    enc_pairs['EncMailsRatioEnc'] = enc_pairs['sent_enc_mails_after_enc_first_mail_from_receiver'] / enc_pairs['sent_mails_after_enc_first_mail_from_receiver'] * 100
    
    print('> enc pairs: \n', enc_pairs[['user', 'sender', 'receiver', 'sent_signed_mails_after_signed_first_mail_from_receiver', 'sent_enc_mails_after_signed_first_mail_from_receiver', 'EncMailsRatio', 'SignedMailsRatio']])
    
    res = analysis_smime_communication_write_result(res, smime_pairs, enc_pairs)

    if onlyUser:
        group = 'user'
    else:
        group = 'sender'
    # non crypto-users:
    non_smime_users = df.groupby(group).filter(lambda x: (x['signed_mails'] == 0).all())
    print('Non smime users:\n ',non_smime_users[[group, 'receiver', 'mails', 'signed_mails']])
    res['countSmimeNotPairUniqueSender'] = "{:,}".format(non_smime_users[group].nunique())
    # count send enc but not signed
    non_smime_users['EncMailsRatio'] = non_smime_users['sent_enc_mails_after_signed_first_mail_from_receiver'] / non_smime_users['sent_mails_after_signed_first_mail_from_receiver'] * 100
    res['noSmimePairs_sent_enc_mails_after_signed_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(smime_pairs['EncMailsRatio'].mean(),2))
    res['noSmimePairs_sent_enc_mails_after_signed_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(smime_pairs['EncMailsRatio'].std(),2))
    
    non_user_send_enc = non_smime_users.loc[(non_smime_users['enc_mails'] > 0)]
    print(non_user_send_enc[['sender', 'receiver', 'mails', 'signed_mails', 'enc_mails']])
    res['noSmimeSenderSendEncMail'] = "{:,}".format(non_user_send_enc['sender'].nunique())
    res['noSmimeUserSendEncMail'] = "{:,}".format(non_user_send_enc['user'].nunique()) 
    res['noSmimeSenderSentEncMailCount'] = "{:,}".format(non_user_send_enc.shape[0])
    res['smimeUserUsesOtherEMailWithoutSmimeToEncrypt'] = "{:,}".format(sum(el in set(non_user_send_enc['user'].tolist()) for el in set(smime_pairs['user'].to_list())))
    res['noSmimeSenderReceivedSignedMail'] ="{:,}".format( non_smime_users[non_smime_users['signed_mails_from_receiver']> 0]['sender'].nunique())
    res['noSmimeUserReceivedSignedMail'] ="{:,}".format( non_smime_users[non_smime_users['signed_mails_from_receiver']> 0]['user'].nunique())
    load_data.write_commands(res,load_data.get_cache_path("numbers_smime_key_exchange.tex", testrun))


def analysis_smime_communication_write_result(res, smime_pairs, enc_pairs):

    res['countSmimeSignPairs'] = "{:,}".format(smime_pairs.shape[0])
    res['countSmimeSignPairUniqueSender'] = "{:,}".format(smime_pairs['sender'].nunique())
    res['countSmimeSignPairUniqueUser'] = "{:,}".format(smime_pairs['user'].nunique())
    
    res['smimeSignPairs_sent_signed_mails_after_signed_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(smime_pairs['SignedMailsRatio'].mean(),2))
    res['smimeSignPairs_sent_signed_mails_after_signed_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(smime_pairs['SignedMailsRatio'].std(),2))
    res['smimeSignPairs_sent_enc_mails_after_signed_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(smime_pairs['EncMailsRatio'].mean(),2))
    res['smimeSignPairs_sent_enc_mails_after_signed_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(smime_pairs['EncMailsRatio'].std(),2))
    
    
    res['smimeSignPairs_sent_signed_mails_after_enc_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(smime_pairs['SignedMailsRatioEnc'].mean(),2))
    res['smimeSignPairs_sent_signed_mails_after_enc_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(smime_pairs['SignedMailsRatioEnc'].std(),2))
    res['smimeSignPairs_sent_enc_mails_after_enc_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(smime_pairs['EncMailsRatioEnc'].mean(),2))
    res['smimeSignPairs_sent_enc_mails_after_enc_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(smime_pairs['EncMailsRatioEnc'].std(),2))
    
    df2 = smime_pairs[['SignedMailsRatio', 'EncMailsRatio', 'SignedMailsRatioEnc', 'EncMailsRatioEnc']]
    df2 = df2.rename(columns={'SignedMailsRatio': 'Signed', 'EncMailsRatio': 'encrypted', 'SignedMailsRatioEnc': 'signed\nafter receiving\nencrypted email', 'EncMailsRatioEnc': 'encrypted\nafter receiving\nencrypted email'})
    plt.clf()
    plt.figure()
    plt.title('Signed email communication partners: \nRatio of signed and encrypted communication')
    df2.boxplot()
    plt.ylabel('Ratio to all emails in %')
    plt.xticks(rotation=45)
    plt.tight_layout()
    plt.savefig(load_data.get_cache_path("graphs/boxplot_key_exchange.png", testrun), bbox_inches='tight')
    # Modin Bug workaround: Use normal pandas instead of modin, otherwise Seaborn crashes
    melted_data = pd.melt(df2._to_pandas())
    graph_support.seaborn_boxplot_paper('sns_boxplot_key_exchange.png', melted_data, x_var="variable", y_var="value", xlabel = None, ylabel = "Ratio to all emails in %", testrun = testrun, hide_legend = True, colors = ["#ecbca7", "#b5caf3", "#ecbca7", "#b5caf3"])

   
    groups = smime_pairs.groupby('sender')
    different_partners = groups['receiver'].count()
    res['partnersPerSenderMean'] = "{:,}".format(round(different_partners.mean(),2))
    res['partnersPerSenderSTD'] = "{:,}".format(round(different_partners.std(),2))
    
    res['countSmimeEncPairs'] = "{:,}".format(enc_pairs.shape[0])
    res['countSmimeEncPairUniqueSender'] = "{:,}".format(enc_pairs['sender'].nunique())
    res['countSmimeEncPairUniqueUser'] = "{:,}".format(enc_pairs['user'].nunique())

    res['smimeEncPairsSentSignedMailsAfterSignedFirstMailFromReceiverRatioMean'] = "{:,}".format(round(enc_pairs['SignedMailsRatio'].mean(),2))
    res['smimeEncPairsSentSignedMailsAfterSignedFirstMailFromReceiverRatioStd'] = "{:,}".format(round(enc_pairs['SignedMailsRatio'].std(),2))
    res['smimeEncPairsSentEncMailsAfterSignedFirstMailFromReceiverRatioMean'] = "{:,}".format(round(enc_pairs['EncMailsRatio'].mean(),2))
    res['smimeEncPairsSentEncMailsAfterSignedFirstMailFromReceiverRatioStd'] = "{:,}".format(round(enc_pairs['EncMailsRatio'].std(),2))
    
    
    res['smimeEncPairs_sent_signed_mails_after_enc_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(enc_pairs['SignedMailsRatioEnc'].mean(),2))
    res['smimeEncPairs_sent_signed_mails_after_enc_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(enc_pairs['SignedMailsRatioEnc'].std(),2))
    res['smimeEncPairs_sent_enc_mails_after_enc_first_mail_from_receiver_ratio_mean'] = "{:,}".format(round(enc_pairs['EncMailsRatioEnc'].mean(),2))
    res['smimeEncPairs_sent_enc_mails_after_enc_first_mail_from_receiver_ratio_std'] = "{:,}".format(round(enc_pairs['EncMailsRatioEnc'].std(),2))
    
    
    df2 = enc_pairs[['SignedMailsRatio', 'EncMailsRatio', 'SignedMailsRatioEnc', 'EncMailsRatioEnc']]
    df2 = df2.rename(columns={'SignedMailsRatio': 'Signed', 'EncMailsRatio': 'encrypted', 'SignedMailsRatioEnc': 'signed\nafter receiving\nencrypted email', 'EncMailsRatioEnc': 'encrypted\nafter receiving\nencrypted email'})
    plt.figure()
    plt.clf()
    box = df2.boxplot()
    plt.xticks(rotation=45)
    plt.title('Encrypted email communication partners: \nRatio of signed and encrypted communication')
    plt.ylabel('Ratio to all emails in %')
    plt.tight_layout()
    plt.savefig(load_data.get_cache_path("graphs/boxplot_key_exchange_enc_pairs.png", testrun), bbox_inches='tight')
   
    groups = enc_pairs.groupby('sender')
    different_partners_enc = groups['receiver'].count()
    res['encPartnersPerSenderMean'] = "{:,}".format(round(different_partners_enc.mean(),2))
    res['encPartnersPerSenderSTD'] = "{:,}".format(round(different_partners_enc.std(),2))
    
    plt.figure()
    plt.clf()
    plt.title('Signed email communication partners: \nPartners per senders')
    plt.ylabel('Absolute number of partners')
    ax1 = different_partners.plot.box()
    plt.savefig(load_data.get_cache_path("graphs/boxplot_key_exchange_partners_sig.png", testrun), bbox_inches='tight')
    graph_support.seaborn_boxplot_paper('sns_boxplot_key_exchange_partners_sig.png', different_partners.to_frame(), x_var="receiver", y_var=None, ylabel = "receiver", xlabel = "Ratio to all emails in %", testrun = testrun, hide_legend = True, ysize=1.25, show_xticks = True)

    
    plt.figure()
    plt.clf()
    plt.title('Encrypted email communication partners: \nPartners per senders')
    plt.ylabel('Absolute number of partners')
    ax2 = different_partners_enc.plot.box()
    plt.savefig(load_data.get_cache_path("graphs/boxplot_key_exchange_partners_enc.png", testrun), bbox_inches='tight')
    return res
    
@cache_it("oliver_smime_all_communication.csv", is_df=True)
def smime_user_all(df = None):
    global testrun
    if df == None:
        df = load_data.load_all_mails(testrun, testrows = 50000)
        df = load_data.drop_error_mails(df)
    df = df[df['cert_data'] != 'empty']
    df = data_support.add_level_info_fix_crypto(df)
    df = df.loc[df['smime_sig_fixed'] == True]
    df = df[['user','sender', 'cert_data']]
    df = df.drop_duplicates()
    
    dict = {'user': [], 'email': [], 'serialnumber': [], 'user issuer': [], 'user issuer country': [], 'root ca': [], 'root country': [], 'ca chain length': [], 'sign algo': [], 'enc algo': [], 'size': [], 'creation date': [], 'expire date': [], 'lifetime': [], 'isExpired': [] ,'certificate': []}
    i = 0
    print("> # cert data: ", df.shape)
    for _, row in df.iterrows():
        res = extract_cert_data(row.values[2])
        res['user'] = row.values[0]
        res['email'] = row.values[1]

        for key, value in res.items():
            dict[key].append(value)
        if i % 500 == 0:
            print("Current row: ", i)
        i += 1
    df_users = pandas.DataFrame(dict)
    print("> Result of smime users at the university: ", df_users.shape)
    df_users['expire date'] = pandas.to_datetime(df_users['expire date']+'0', format='%W:%Y%w')
    df_users['creation date'] = pandas.to_datetime(df_users['creation date']+'0', format='%W:%Y%w')
    return df_users
    
    
@cache_it("oliver_smime_user_sent_certificates.csv", is_df=True)
def smime_user_university(df = None):
    """
        input: a dataframe with all mails
        output: a dataframe with all sent certificates from users of the university
        The dataframe looks like:
        user | sender | serialnumber | creation | expire | user issuer | root ca | json data
    """
    global testrun
    if df == None:
        df = load_data.load_all_mails(testrun, testrows = 5000000)
        df = load_data.drop_error_mails(df)
    df = df[df['cert_data'] != 'empty']
    df = data_support.add_level_info_fix_crypto(df)
    df = df.loc[df['smime_sig_fixed'] == True]
    df = data_support.get_sent_mails(df)
    df = df[['user','sender', 'cert_data']]
    df = df.drop_duplicates()

   
    dict = {'user': [], 'email': [], 'serialnumber': [], 'user issuer': [], 'user issuer country': [], 'root ca': [], 'root country': [], 'ca chain length': [], 'sign algo': [], 'enc algo': [], 'size': [], 'creation date': [], 'expire date': [], 'lifetime': [], 'isExpired': [] ,'certificate': []}
    i = 0
    for _, row in df.iterrows():
        res = extract_cert_data(row.values[2])
        res['user'] = row.values[0]
        res['email'] = row.values[1]

        for key, value in res.items():
            dict[key].append(value)
        if i % 500 == 0:
            print("Current row: ", i)
        i += 1
    df_users = pandas.DataFrame(dict)
    print("> Result of smime users at the university: ", df_users.shape)
    df_users['expire date'] = pandas.to_datetime(df_users['expire date']+'0', format='%W:%Y%w')
    df_users['creation date'] = pandas.to_datetime(df_users['creation date']+'0', format='%W:%Y%w')
    return df_users.drop_duplicates(['user', 'email', 'serialnumber', 'user issuer'])
    
def extract_cert_data(cert):
    org = cert
    cert = json.loads(cert)
    cert_record = {}
    length = len(cert)
    if length > 0:
        root_cert = cert[0]
        analysis_root_ca(root_cert, cert_record)
        i = 1
        user_cert = cert[len(cert)-i]
        # Often the last entry is the level, e.g. {'level': 3}.
        while not ('sn' in user_cert):
            i += 1
            user_cert = cert[len(cert)-i]
        length = length - i + 1
        analysis_user_cert(user_cert, cert_record)
    cert_record['ca chain length'] = length
    cert_record['certificate'] = org
    return cert_record
  
@cache_it("oliver_smime_certificate_rollover.csv", is_df=True)
def certificate_rollover(df2, test):
    """
        Output: dataframe of rollover
        
        Definition of rollover:
            Let cert_1 and cert_2 be certificates with creation date cd_1 resp. cd_2 and expire date ed_1 resp. ed_2.
            We say certificate rollover if and only if cd_1 < cd_2 and ed_1 < ed_2.
            In this case, the user introduces a new key beyond the lifetime of a previous key.
            If cd_1 < cd_2 and ed_2 <= ed_1 then it is no key rollover because the new key is not longer valid compared to the previous key.
            
            A key rollover has a gap if and only if ed_1 < cd_2. Here is a period of time where no key is valid.
        
        cd = creation date, ed = expire date
        E.g.
        cd_1 , ed_1, cd_2, ed_2 -> key rollover -> gap between key 1 and key 2
        cd_1, cd_2, ed_1, ed_2 -> key rollover -> no gap between key 1 and key 2
        cd_1, cd_2, cd_3, ed_3, ed_1, ed_2 -> key rollover between key 1 and 2 but not with key 3 -> no gap
        cd_1, cd_2, ed_1, cd_3, ed_3, ed_2 -> key rollover between key 1 and 2 but not with key 3 -> no gap
        
        output
        'user', 'email', 'from','from user issuer', 'to', 'to user issuer', 'gap', 'from_creation', 'from_expire', 'to_creation', 'to_expire', 'type'
        
    """
    i = 0
    df = pandas.DataFrame(columns=['user', 'email', 'from','from user issuer', 'to', 'to user issuer', 'gap', 'from_creation', 'from_expire', 'to_creation', 'to_expire', 'type'])
    
    
    print(df2)
    #df2['expire date'] = pandas.to_datetime(df2['expire date']+'0', format='%W:%Y%w')
    #df2['creation date'] = pandas.to_datetime(df2['creation date']+'0', format='%W:%Y%w')
    df2 = df2.groupby('email')
    print("> start loop:", len(df2))
    for name, group in df2:
        rollovers = certificate_rollover_algo(group)
        if i % 10 == 0:
            print("> current column", i, "#rollovers: ", len(rollovers), "Group size: ", group.shape)
        for roll in rollovers:
            df = df.append(roll, ignore_index = True)
        i += 1
        
    return df

def certificate_rollover_algo(certs):
    if certs.shape[0] < 2:
        # Short cut we do not have any roll over...
        return []
    if isinstance(certs['expire date'], str):
        certs['expire date'] = pandas.to_datetime(certs['expire date']+'0', format='%W:%Y%w')
    if isinstance(certs['creation date'], str):
        certs['creation date'] = pandas.to_datetime(certs['creation date']+'0', format='%W:%Y%w')
        
    my_certs = []
    for i in range(0, certs.shape[0]):
        cert = certs.iloc[i]
        cert['creation date'] = pandas.to_datetime(cert['creation date'], format='%Y-%m-%d')
        cert['expire date'] = pandas.to_datetime(cert['expire date'], format='%Y-%m-%d')
        my_certs.append(cert)
    certs = my_certs
    if len(certs) <= 1:
        return []
    rollovers = []
   
    certs = sorted(certs, key=lambda k: k['creation date'])
    
    # current cert is the first created one
    current_cert = certs[0]
    del certs[0]
    
    for c in certs:
        if c['creation date'] == current_cert['creation date']:
            # No rollover?!
            # <-----current cert----->
            # <----new cert----> or
            # <-----new cert---------> (same expire date) or
            # <-------------new cert--------->
            type = 'dual creation'
        elif c['creation date'] <= current_cert['expire date'] and c['expire date'] > current_cert['expire date']:
            # Rollover with a good timing
            # <-----current cert----->
            #                <--------new cert---->
            type = 'safe'
            
        elif c['creation date'] > current_cert['expire date'] and c['expire date'] > current_cert['expire date']:
            # rollover but to late
            # <-----current cert----->
            #                           <--------new cert---->
           type = 'late'
        elif c['expire date'] <= current_cert['expire date']:
            # only a temporary rollover
            # <------------current cert----------->
            #       <----new cert-------->
            type = 'temp'
        gap = (c['creation date'] - current_cert['expire date']) / np.timedelta64(1, 'D')
        rollover = {'user': c['user'], 'email': c['email'], 'from': current_cert['serialnumber'], 'to': c['serialnumber'], 'gap': gap,'from_creation': current_cert['creation date'], 'from_expire': current_cert ['expire date'], 'to_creation': c['creation date'], 'to_expire': c['expire date'], 'type': type, 'from user issuer': current_cert['user issuer'], 'to user issuer': c['user issuer']}
        if rollover['from'] != rollover['to']:
            rollovers.append(rollover)
        # Current cert depends on the expire date
        if c['expire date'] > current_cert['expire date']:
            current_cert = c
    return rollovers

def analysis_user_cert(cert, res):
    key_pairs = [   ('sn', 'serialnumber'),
                    ('algo', 'sign algo'),
                    ('size', 'size'),
                    ('type', 'enc algo'),
                    ('not_valid_b', 'creation date'),
                    ('not_valid_a', 'expire date')]
    for pair in key_pairs:
        cert_key = pair[0]
        res_key = pair[1]
        if cert_key in cert:
            res[res_key] = cert[cert_key]
        else:
            res[res_key] = 'None'
    res['lifetime'] = data_support.time_distance(res['creation date'], res['expire date'])
    res ['isExpired'] = True
    if res['expire date'][-4:] > '2020':
        res ['isExpired'] = False
    if "issuer" in cert:
        issuer = cert["issuer"]
        if "commonName" in issuer:
            res['user issuer'] = issuer["commonName"]
        elif "organizationalUnitName" in issuer:
            res['user issuer'] = issuer["organizationalUnitName"]
        elif "organizationName" in issuer:
            res['user issuer'] = issuer["organizationName"]
        else:
            res['user issuer'] = "None"
            
        if "countryName" in issuer:
            res['user issuer country'] = issuer["countryName"]
        else:
            res['user issuer country'] = "None"
    else:
        res['user issuer'] = "None"
        res['user issuer country'] = "None"
    return res

def smime_pgp_key_table():
    df = smime_user_all()
    smime = analysis_certificates(df)
    total = smime['unique_keys']
    res = {}
    res['totalSMIME'] = "{:,}".format(total)
    oddnumbers = {}
    for name, count in smime['enc_algo'].items():
        if(str(name[1]) not in ("256", "521", "768", "1024", "2048", "3072", "4096", "8192")):
            if name[0] not in oddnumbers:
                oddnumbers[name[0]] = 0
            oddnumbers[name[0]]+=count
        name = 'SMIME-' + name[0] + '-' + str(name[1])
        value = "{:,}".format(count) + " ({:,}\%)".format(round(100.0 * count / total, 2))
        res[name] = value

    for odd in oddnumbers:
        value = "{:,}".format(oddnumbers[odd]) + " ({:,}\%)".format(round(100.0 * oddnumbers[odd] / total, 2))
        res["SMIME-"+odd+"-odd"] = value

    load_data.write_var(res, load_data.get_cache_path("numbers_key_table_smime.tex", testrun))


def analysis_root_ca(root_cert, res):
    if "issuer" in root_cert:
        issuer =  root_cert["issuer"]
        if "commonName" in issuer:
            res["root ca"] = issuer["commonName"]
        elif "organizationalUnitName" in issuer:
            res["root ca"] = issuer["organizationalUnitName"]
        elif "organizationName" in issuer:
            res["root ca"] = issuer["organizationName"]
        else:
            res["root ca"] = "None"
            
        if "countryName" in issuer:
            res["root country"] = issuer["countryName"]
        else:
            res["root country"] = "None"
    else:
        res["root ca"] = "None"
        res["root country"] = "None"
    return res



def analysis_certificates(smime_users = None, fileName = None):
    """
    
      Data: user | email | serialnumber | creation | expire | user issuer | root ca | json data
      Questions:
        * Number of certificates_ Done
        * Summary of CAs: Done
        * Certificates per email -> mean, std, max, ones,
        * Certificates with SHA-1 -> Done (max creation/expire date)
        * Signatur algo/ enc key/size -> Done
        
    Output:
    A dictionary with general information:
        * unique_keys -> number of unique certificates
        * sign_algo -> a dictionary with hash algo + number of related certificates
        * enc_algo -> a dictionary with enc algo + number of related certificates
        * enc_key_size -> a dictionary with encryption key size + number of related certificates
        * liftetime -> a dictionary with the description of the valid time period (mean, min, max, std, ...)
    """
    pandas.set_option('display.max_rows', 5500)
    pandas.set_option('display.max_columns', 10)
    pandas.set_option('display.width', 300)

    global YEAR
    
    only_sent_folder = False
    if not isinstance(smime_users, pandas.DataFrame):
        only_sent_folder = True
        smime_users = smime_user_university()
        print(">only university!")
    res = {}
    
    table = {}
    
    res['distinctSmimeAddressCount'] = smime_users['email'].nunique()
    if 'empty' in smime_users['email'].unique():
        res['distinctSmimeAddressCount'] = res['distinctSmimeAddressCount'] - 1
        res['distinctSmimeAddressCount'] = "{:,}".format(res['distinctSmimeAddressCount'])
    res['distinctRootCaCount'] = "{:,}".format(smime_users['root ca'].nunique() - 1) # We count None too...
    res['distinctIssuerCount'] = "{:,}".format(smime_users['user issuer'].nunique() -1) # We count None too...
    

    unique_certs = smime_users[['serialnumber', 'user issuer', 'user issuer country', 'root ca', 'root country', 'ca chain length','sign algo', 'enc algo', 'size', 'creation date', 'expire date', 'lifetime', 'isExpired']].drop_duplicates(subset=['serialnumber', 'user issuer'])
    res['NoIssuerCount'] = "{:,}".format(unique_certs.loc[(unique_certs['user issuer'].isin(['None']))].shape[0])
    res['NoIssuerCountPercent'] = "{:,}".format(round(100.00*unique_certs.loc[(unique_certs['user issuer'].isin(['None']))].shape[0] / unique_certs.shape[0],2))
    res['NoRootCount'] = "{:,}".format(unique_certs.loc[(unique_certs['root ca'].isin(['None']))].shape[0])
    res['NoRootCountPercent'] = "{:,}".format(round(100.00*unique_certs.loc[(unique_certs['root ca'].isin(['None']))].shape[0] / unique_certs.shape[0],2))
    
    res['distinctSmimeCertificatesCount'] = "{:,}".format(unique_certs.shape[0])
    table['unique_keys'] = unique_certs.shape[0]
    issuer = unique_certs[['user issuer']].value_counts().to_dict()
    res, issuerWith1Cert, issuerWith2Certs = analysis_certificates_issuer(res, issuer)
    
    # Look on chain length issuers with only one or two certificates
    issuerWith1CertNoRoot = unique_certs.loc[(unique_certs['user issuer'].isin(issuerWith1Cert)) & (unique_certs['ca chain length'] < 3)]
    res['IssuerWithOnlyOneButNoRoot'] = "{:,}".format(issuerWith1CertNoRoot.shape[0])
    issuerWith2CertsNoRoot = unique_certs.loc[(unique_certs['user issuer'].isin(issuerWith2Certs))& (unique_certs['ca chain length'] < 3)]
    res['IssuerWithOnlyTwoButNoRoot'] = "{:,}".format(issuerWith2CertsNoRoot.shape[0])
    res['IssuerWithOnlyOneOrTwoButNoRoot'] = "{:,}".format(issuerWith1CertNoRoot.shape[0] + issuerWith2CertsNoRoot.shape[0])
    res['IssuerWithOnlyOneOrTwoButNoRootPercent'] = "{:,}".format(round(100.00 * (issuerWith1CertNoRoot.shape[0] + issuerWith2CertsNoRoot.shape[0]) /(len(issuerWith1Cert) + len(issuerWith2Certs)), 2))

    res = analysis_certificates_algos(res, table, unique_certs)
    res = analysis_certificates_root(res, unique_certs['root ca'].value_counts().to_dict())
    
    chain = unique_certs['ca chain length'].value_counts().to_dict()

    res['ChainLengthOne'] = "{:,}".format(chain[1])
    res['ChainLengthOnePercent'] = "{:,}".format(round(chain[1] / unique_certs.shape[0] * 100, 2))

    res['smimeLifetimeInYearMean'] = "{:,}".format(round(unique_certs['lifetime'].mean() / YEAR, 2))
    res['smimeLifetimeInYearMedian'] = "{:,}".format(round(unique_certs['lifetime'].median() / YEAR, 2))

    res['smimeLifetimeInYearSTD'] = "{:,}".format(round(unique_certs['lifetime'].std() / YEAR, 2))
    res['smimeLifetimeInYearMax'] = "{:,}".format(round(unique_certs['lifetime'].max() / YEAR, 2))
    res['smimeLifetimeInWeeksMin'] = "{:,}".format(round(unique_certs['lifetime'].min() / 7,2))
    
    table['smimeLifetimeInYearMean'] = res['smimeLifetimeInYearMean']
    table['smimeLifetimeInYearSTD'] = res['smimeLifetimeInYearSTD']
    table['smimeLifetimeInYearMax'] = res['smimeLifetimeInYearMax']
    table['smimeLifetimeInWeeksMin'] = res['smimeLifetimeInWeeksMin']
    
    res['smimeCertExpiredCount'] = "{:,}".format(unique_certs['isExpired'].sum())
    res['smimeCertExpiredPercent'] = "{:,}".format(round(100.0 * unique_certs['isExpired'].sum() / unique_certs.shape[0], 2))
    
    res['smimeCertCreatedTwentyPercent'] = "{:,}".format(round(100.00 * unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2020'].shape[0] /unique_certs.shape[0],2))
    res['smimeCertCreatedNineteenPercent'] = "{:,}".format(round(100.00 * unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2019'].shape[0] /unique_certs.shape[0],2))
    res['smimeCertCreatedEightteenPercent'] = "{:,}".format(round(100.00 * unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2018'].shape[0] /unique_certs.shape[0],2))
    res['smimeCertCreatedSventeenPercent'] = "{:,}".format(round(100.00 * unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2017'].shape[0] /unique_certs.shape[0],2))
    res['smimeCertCreatedLastFiveYears'] = "{:,}".format(round(100.00 * unique_certs[(unique_certs['creation date'].astype(str).str.slice(0, 4) == '2020') | (unique_certs['creation date'].astype(str).str.slice(0, 4) == '2019') | (unique_certs['creation date'].astype(str).str.slice(0, 4) == '2018') | (unique_certs['creation date'].astype(str).str.slice(0, 4) == '2017') | (unique_certs['creation date'].astype(str).str.slice(0, 4) == '2016') | (unique_certs['creation date'].astype(str).str.slice(0, 4) == '2015')].shape[0] /unique_certs.shape[0],2))
    res['smimeCertCreatedTwentyCount'] = "{:,}".format(round(unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2020'].shape[0]))
    res['smimeCertCreatedNineteenCount'] = "{:,}".format(round(unique_certs[unique_certs['creation date'].astype(str).str.slice(0, 4) == '2019'].shape[0]))

    if only_sent_folder:
        certs_per_email = smime_users.loc[smime_users['email'] != 'empty']
        certs_per_email = certs_per_email['email'].value_counts()
        analysis_certificates_emails(res, certs_per_email)
    
    print("> results of S/MIME certs")
    
    print(unique_certs['lifetime'].describe())
    print('Creation date: \n', unique_certs['creation date'].astype(str).str.slice(0, 4).value_counts())
    print(unique_certs['creation date'].astype(str).str.slice(0, 4).describe())
    for k, v in res.items():
        print(k, ':', v)
    
    if fileName != None:
        load_data.write_commands(res, load_data.get_cache_path("numbers_smime_certificates"+fileName+".tex", testrun))
    return table

def addressHasUser(addresses):
    df = load_data.load_all_mails(testrun, columns =['user', 'sender', 'is_sent'])
    df = data_support.get_sent_mails(df)
    df = df.drop_duplicates()
    result = []
    for address in addresses:
        res = df.loc[df['sender'] == address]
        result.append(res.shape[0])
        print(address, res.shape[0])
    return result
    
def analysis_certificates_algos(res, table, unique_certs):
    """
    Known sign algos:  {'sha256': 6424, 'sha1': 2116, 'md5': 166, 'sha512': 88, 'sha384': 18, '1.2.840.113549.1.1.10': 4, 'md2WithRSAEncryption': 1}
    UserWarning: `Series.to_dict` defaulting to pandas implementation.
    Known enc algos:  {'RSA': 8815, 'EC': 2}
    {2048: 8161, 4096: 353, 1024: 264, 3072: 26, 512: 6, 8192: 3, 256: 1, 768: 1, 1023: 1, 521: 1}
    Known key size 2048    8161
    4096     353
    1024     264
    3072      26
    512        6
    8192       3
    256        1
    768        1
    1023       1
    521        1
    dtype: int64

    """

    sign_algos = unique_certs['sign algo'].value_counts().to_dict()
    table['sign_algos'] = sign_algos
    
    res['signAlgoSHAOne'] = "{:,}".format(sign_algos['sha1'])
    res['signAlgoSHAOnePercent'] = "{:,}".format(round(100.0*sign_algos['sha1']/unique_certs.shape[0]))
    
    res['signAlgoSHATwoFiveSix'] = "{:,}".format(sign_algos['sha256'])
    res['signAlgoSHATwoFiveSixPercent'] = "{:,}".format(round(100.0*sign_algos['sha256']/unique_certs.shape[0]))
    if 'sha512' in sign_algos:
        res['signAlgoSHAFiveOneTwo'] = "{:,}".format(sign_algos['sha512'])
        res['signAlgoSHAFiveOneTwoPercent'] = "{:,}".format(round(100.0*sign_algos['sha512']/unique_certs.shape[0]))
    if 'sha384' in sign_algos:
        res['signAlgoSHAThreeEightFour'] = "{:,}".format(sign_algos['sha384'])
        res['signAlgoSHAThreeEightFourPercent'] = "{:,}".format(round(100.0*sign_algos['sha384']/unique_certs.shape[0]))
    if 'md5' in sign_algos:
        res['signAlgoMDFive'] = "{:,}".format(sign_algos['md5'])
        res['signAlgoMDFivePercent'] = "{:,}".format(round(100.0*sign_algos['md5']/unique_certs.shape[0]))
    
    sha1 = unique_certs.loc[unique_certs['sign algo'] == 'sha1']
    sha1.sort_values(by=['creation date'], inplace=True)
    res['signSHAOneLastCreationDate'] = sha1['creation date'].max()
    res['signSHAOneLastExpireDate'] = sha1['expire date'].max()
    print('SHA-1 creation dates and type:\n', sha1['creation date'].value_counts(), type(sha1['expire date'].max()))
    current_year = sha1[sha1['creation date'].astype(str).str.slice(0, 4) == '2020']
    print('SHA1 in 2020:\n', current_year)
    res['signSHAOneLastYearCount'] = "{:,}".format(current_year.shape[0])

    
    md5 = unique_certs.loc[unique_certs['sign algo'] == 'md5']
    md5.sort_values(by=['creation date'], inplace=True)
    res['signMDFiveCreationDate'] = md5['creation date'].max()
    res['signMDFiveLastExpireDate'] = md5['expire date'].max()
    print('MD5:\n', md5)

    print('> known sign algos: ', sign_algos)
    enc_algos = unique_certs['enc algo'].value_counts().to_dict()
    
    print('> known enc algos: ', enc_algos)
    res['encAlgoRSA'] = enc_algos['RSA']
    if 'EC' in enc_algos:
        res['encAlgoEC'] = enc_algos['EC']

    key_size = unique_certs['size'].value_counts().to_dict()
    
    table['enc_algo'] = unique_certs.groupby(['enc algo', 'size']).size()
        
    print('> key sizes: \n', key_size)
    if 2048 in key_size:
        res['encAlgoSizeTwokPercent'] = "{:,}".format(round(key_size[2048]/res['encAlgoRSA']*100,2))
    if 3072 in key_size:
        res['encAlgoSizeThreekPercent'] = "{:,}".format(round(key_size[3072]/res['encAlgoRSA']*100,2))
    if 4096 in key_size:
        res['encAlgoSizeFourkPercent'] = "{:,}".format(round(key_size[4096]/res['encAlgoRSA']*100,2))
    if 1024 in key_size:
        res['encAlgoSizeOnekPercent'] = "{:,}".format(round(key_size[1024]/res['encAlgoRSA']*100,2))
        res['encAlgoSizeOnek'] = "{:,}".format(key_size[1024])
        small = unique_certs.loc[unique_certs['size'] == 1024]
        small.sort_values(by=['creation date'], inplace=True)
        res['encAlgoSizeOnekLastCreationDate'] = small['creation date'].max()
        print('Value/Type of current year: ', small['creation date'].max(), type(small['creation date'].max()))
        current_year = small[small['creation date'].astype(str).str.slice(0, 4) == small['creation date'].max()[0: 4]]
        res['encAlgoSizeOnekLastCreationDateCount'] = "{:,}".format(current_year.shape[0])
        
    if 512 in key_size:
        res['encAlgoSizeFiveOneTwoPercent'] = "{:,}".format(round(key_size[512]/res['encAlgoRSA']*100,2))
        res['encAlgoSizeFiveOneTwo'] = "{:,}".format(key_size[512])
        small = unique_certs.loc[unique_certs['size'] == 512]
        small.sort_values(by=['creation date'], inplace=True)
        print('Key size 512:\n', small)
        res['encAlgoSizeFiveOneTwoLastCreationDate'] = small['creation date'].max()
        print('Value/Type of current year: ', small['creation date'].max(), type(small['creation date'].max()))
        current_year = small[small['creation date'].astype(str).str.slice(0, 4) == small['creation date'].max()[0: 4]]
        res['encAlgoSizeFiveOneTwokLastCreationDateCount'] = "{:,}".format(current_year.shape[0])
    if 8192 in key_size:
        res['encAlgoSizeEightkPercent'] = "{:,}".format(round(key_size[8192]/res['encAlgoRSA']*100,2))
        
    res['encAlgoRSA'] = "{:,}".format(res['encAlgoRSA'])
    print('> known key size', unique_certs['size'].value_counts())
    
    print('> over all Creation dates:\n', unique_certs['creation date'].value_counts())

    return res

def analysis_certificates_emails(res, certs_per_email):
    res['certsPerEMailMean'] = "{:,}".format(round(certs_per_email.mean(),2))
    res['certsPerEMailSTD'] = "{:,}".format(round(certs_per_email.std(),2))
    res['oneCertPerEMail'] = "{:,}".format(certs_per_email[certs_per_email == 1].size)
    res['moreCertsPerEMail'] = "{:,}".format(certs_per_email[certs_per_email > 1].size)
    res['maxCertsPerEMail'] = "{:,}".format(certs_per_email.max()+0)
    suspicous = certs_per_email[certs_per_email > 10].index.tolist()
    return res

def analysis_certificates_root(res, root):
    # Counts DFN certs and others
    # Depends on data set!
    one = 0
    two = 0
    total = root['None']
    root_summary = root
    telekom = 0
    dfn = 0
    
    del root['None']
    
    for key, value in root.items():
        total += value
        if 'Telekom' in key:
            telekom += value
        elif 'T-TeleSec' in key:
            telekom += value
        if 'DFN' in key:
            dfn += value
        if value == 1:
            one += 1
        elif value == 2:
            two += 1
    res['RootWithOnlyOneCert'] = "{:,}".format(one)
    res['RootWithOnlyTwoCert'] = "{:,}".format(two)
    res['TotalRootCertificates'] = "{:,}".format(total)
    
    res['RootTelekomPercent']  = "{:,}".format(round(telekom/total * 100, 2))
    res['RootTelekom']  = "{:,}".format(telekom)
    res['RootDFNPercent']  = "{:,}".format(round(dfn/total * 100, 2))
    res['RootDFN'] = "{:,}".format(dfn)
    
    issuer_summary= dict(sorted(root_summary.items(), key=lambda item: item[1], reverse=True))
    i = 0
    others = total
    for key, value in root_summary.items():
        if i == 0:
            res['FirstRoot'] = key
            res['FirstRootCountPercent'] = "{:,}".format(round(value/total*100, 2))
            others = others - value
        elif i == 1:
            res['SecondRoot'] = key
            res['SecondRootCountPercent'] = "{:,}".format(round(value/total*100, 2))
            others = others - value
        elif i == 2:
            res['ThirdRoot'] = key
            res['ThirdRootCountPercent'] = "{:,}".format(round(value/total*100, 2))
            others = others - value
        else:
            res['OtherRootCountPercent'] = "{:,}".format(round(others/total*100, 2))
            break
        i += 1
    return res


def analysis_certificates_issuer(res, issuer):
    # Counts DFN certs and others
    # Depends on data set!
    one = 0
    two = 0
    total = 0
    issuer_summary = {'DFN': 0, 'Leuphana': 0, 'COMODO': 0, 'LUH': 0, 'TERENA': 0, 'TU Dresden': 0, 'SwissSign': 0, 'MPG': 0, 'TU Clausthal': 0, 'KIT': 0}
    issuerWith1Cert = []
    issuerWith2Certs = []
    for key, value in issuer.items():
        key = key[0]
        total += value
        if 'DFN' in key:
            issuer_summary['DFN'] += value
            print(key)
            if '1' in key:
                if 'DFNCertGOne' in res:
                    res['DFNCertGOne'] += value
                else:
                    res['DFNCertGOne'] = value
            elif '2' in key:
                if 'DFNCertGTwo' in res:
                    res['DFNCertGTwo'] += value
                else:
                    res['DFNCertGTwo'] = value
        elif 'Leuphana' in key:
            issuer_summary['Leuphana'] += value
        elif 'COMODO' in key:
            issuer_summary['COMODO'] += value
        elif 'LUH' in key:
            issuer_summary['LUH'] += value
        elif 'TERENA' in key:
            issuer_summary['TERENA'] += value
        elif 'Dresden' in key:
            issuer_summary['TU Dresden'] += value
        elif 'SwissSign' in key:
            issuer_summary['SwissSign'] += value
        elif 'MPG' in key:
            issuer_summary['MPG'] += value
        elif 'Clausthal' in key:
            issuer_summary['TU Clausthal'] += value
        elif 'KIT' in key:
            issuer_summary['KIT'] += value
        else:
            if 'None' in key:
                print("NONE: ", key)
            else:
                issuer_summary[key] = value
        if value == 1:
            issuerWith1Cert.append(key)
            one += 1
        elif value == 2:
            issuerWith2Certs.append(key)
            two += 1
    res['IssuerWithOnlyOneCert'] = "{:,}".format(one)
    res['IssuerWithOnlyTwoCert'] = "{:,}".format(two)
    res['TotalIssuedCertificates'] = "{:,}".format(total)
    issuer_summary= dict(sorted(issuer_summary.items(), key=lambda item: item[1], reverse=True))
    i = 0
    others = total
   
    for key, value in issuer_summary.items():
        if i == 0:
            res['FirstIssuer'] = key
            res['FirstIssuerCount'] = "{:,}".format(value)
            res['FirstIssuerCountPercent'] = "{:,}".format(round(100.00*value/total, 2))
            others = others - value
        elif i == 1:
            res['SecondIsssuer'] = key
            res['SecondIssuerCount'] = "{:,}".format(value)
            res['SecondIssuerCountPercent'] = "{:,}".format(round(100.00*value/total, 2))
            others = others - value
        elif i == 2:
            res['ThirdIssuer'] = key
            res['ThirdIssuerCount'] = "{:,}".format(value)
            res['ThirdIssuerCountPercent'] = "{:,}".format(round(100.00*value/total, 2))
            others = others - value
        else:
            res['OtherCount'] = "{:,}".format(others)
            res['OtherCountPercent'] = "{:,}".format(round(100.00*others/total, 2))
            break
        i += 1
    return res, issuerWith1Cert, issuerWith2Certs
    
def analysis_multiple_keys():
    global YEAR
    res = {}
    pandas.set_option('display.max_rows', 5500)
    pandas.set_option('display.max_columns', 12)
    pandas.set_option('display.width', 500)
    pandas.set_option('display.max_colwidth', -1)
    """
    {'user': [], 'email': [], 'serialnumber': [], 'user issuer': [], 'user issuer country': [], 'root ca': [], 'root country': [], 'ca chain length': [], 'sign algo': [], 'enc algo': [], 'size': [], 'creation date': [], 'expire date': [], 'lifetime': [], 'isExpired': [] ,'certificate': []}
    """
    smime_users = smime_user_university().drop_duplicates(subset=['email', 'serialnumber', 'user issuer'])
    issuer_isDFN(smime_users)
    # keys per email
    keys_per_emails = smime_users.groupby('email')
    count_keys_per_emails = keys_per_emails.size()
    print(count_keys_per_emails.describe())
    res['smimeKeysPerEMailMean'] = "{:,}".format(round(count_keys_per_emails.mean(), 2))
    res['smimeKeysPerEMailSTD'] = "{:,}".format(round(count_keys_per_emails.std(), 2))
    res['smimeKeysPerEMailMax'] = "{:,}".format(count_keys_per_emails.max())
    
    # active vs inactive emails
    active = smime_users.loc[smime_users['isExpired'] == False].groupby('email')
    res['smimeEMailCount'] = "{:,}".format(count_keys_per_emails.count())
    res['smimeActiveEMailCount'] = "{:,}".format(active.size().count())
    res['smimeActiveEMailCountPercent'] = "{:,}".format(round(100 * active.size().count() / count_keys_per_emails.count(), 2))
    res['smimeInactiveEMailCount'] = "{:,}".format(count_keys_per_emails.count() - active.size().count())
    res['smimeInactiveEMailCountPercent'] = "{:,}".format(round(100 * (count_keys_per_emails.count() - active.size().count()) / count_keys_per_emails.count(),2))

    # emails with more than two keys
    multi_key_emails = keys_per_emails.filter(lambda x: len(x) > 1)
    count_multi_key_emails = multi_key_emails.groupby('email').size()
    print(count_multi_key_emails.describe())
    res['smimeMultipleKeysEmailCount'] = "{:,}".format(count_multi_key_emails.count())
    res['smimeMultipleKeysEmailCountPercent'] = "{:,}".format(round(100 * count_multi_key_emails.count() / count_keys_per_emails.count(), 2))
    res['smimeMultipleKeysEmailKeysMean'] = "{:,}".format(round(count_multi_key_emails.mean(),2))
    res['smimeMultipleKeysEmailKeysSTD'] = "{:,}".format(round(count_multi_key_emails.std(),2))
        
    res['smimeMultipleKeysEmailLifeTimeInYearMean'] = "{:,}".format(round(multi_key_emails['lifetime'].mean() / YEAR, 2))
    res['smimeMultipleKeysEmailLifeTimeSTD'] = "{:,}".format(round(multi_key_emails['lifetime'].std() / YEAR, 2))
    res['smimeMultipleKeysIssuerCountMean'] = "{:,}".format(round((multi_key_emails.groupby('email')['user issuer'].nunique()).mean(), 2))
    res['smimeMultipleKeysIssuerCountSTD'] = "{:,}".format(round((multi_key_emails.groupby('email')['user issuer'].nunique()).std(), 2))
    res['smimeMultipleKeysRootCountMean'] = "{:,}".format(round((multi_key_emails.groupby('email')['root ca'].nunique()).mean(), 2))
    res['smimeMultipleKeysRootCountSTD'] = "{:,}".format(round((multi_key_emails.groupby('email')['root ca'].nunique()).std(), 2))
    
    res['smimeMultipleKeysEmailExpiredCount'] = "{:,}".format(count_multi_key_emails.count() - multi_key_emails.loc[multi_key_emails['isExpired'] == False].groupby('email').size().count())
    res['smimeMultipleKeysEmailExpiredCountPercent'] = "{:,}".format(round(100.00 * (count_multi_key_emails.count() - multi_key_emails.loc[multi_key_emails['isExpired'] == False].groupby('email').size().count()) / count_multi_key_emails.count(), 2))
    # Change between DFN
    dfn_email = multi_key_emails.loc[multi_key_emails['isDFN-LUH'] == False].groupby('email').size()
    res['smimeMultipleKeysOnlyDFNLUHEmailsCount'] = "{:,}".format(count_multi_key_emails.count() - dfn_email.count())
    res['smimeMultipleKeysOnlyDFNLUHEmailsCountPercent'] = "{:,}".format(round(100 * (count_multi_key_emails.count() - dfn_email.count()) / count_multi_key_emails.count(),2))


    # Single keys
    single_key_emails = keys_per_emails.filter(lambda x: len(x) == 1)
    res['smimeSingleKeyEmailLifeTimeMean'] = "{:,}".format(round(single_key_emails['lifetime'].mean() / YEAR, 2))
    res['smimeSingleKeyEmailLifeTimeSTD'] = "{:,}".format(round(single_key_emails['lifetime'].std() / YEAR, 2))
    res['smimeSingleKeyEmailExpiredCount'] = "{:,}".format(single_key_emails.loc[single_key_emails['isExpired'] == False].shape[0])
    res['smimeSingleKeyEmailExpiredCountPercent'] = "{:,}".format(round(100.00 * single_key_emails.loc[single_key_emails['isExpired'] == False].shape[0] / single_key_emails.shape[0], 2))

    load_data.write_commands(res, load_data.get_cache_path("numbers_smime_multiple_certificates.tex", testrun))
    


def issuer_isDFN(df):
    df['isDFN-LUH'] = df['user issuer'].str.contains('DFN') | df['user issuer'].str.contains('LUH')
    
def analysis_rollover(test):
    """
        Gap in weeks!
    """
    smime_users = smime_user_university()
    data = []
    df = certificate_rollover(smime_users, test)
    df['gap'] = df['gap'] /7
    print("> rollovers: \n", df[['user', 'type', 'from', 'to', 'gap']])
    
    pandas.set_option('display.max_rows', 5500)
    pandas.set_option('display.max_columns', 12)
    pandas.set_option('display.width', 500)
    pandas.set_option('display.max_colwidth', -1)

    df['gap'] = df['gap'].astype(int)
  
    groups = df.groupby('type', group_keys=True)['gap']
    res = {}
    res['SMIMEPossibleRolloverCount'] = df.shape[0]
    week4 = groups['gap'].apply(lambda x: (x <= 4).sum()).reset_index(name='count')
    
    print('<=4 weeks: ', week4.loc[week4['type'].str.contains('late')])
    print('<=8 weeks: ', groups['gap'].apply(lambda x: (x <= 8).sum()).reset_index(name='count'))
    print('<=52 weeks: ', groups['gap'].apply(lambda x: (x <= 52).sum()).reset_index(name='count'))
    print('>52 weeks: ', groups['gap'].apply(lambda x: (x > 52).sum()).reset_index(name='count'))

    print('>= -4 weeks:', groups['gap'].apply(lambda x: (x >= -4).sum()).reset_index(name='count'))
    print('>= -8 weeks:', groups['gap'].apply(lambda x: (x >= -8).sum()).reset_index(name='count'))
    print('>= -52 weeks:', groups['gap'].apply(lambda x: (x >= -52).sum()).reset_index(name='count'))
    print('< -52 weeks:', groups['gap'].apply(lambda x: (x < -52).sum()).reset_index(name='count'))
    
    if 'late' in groups.indices:
        res['SMIMERolloverLateCount'] = "{:,}".format(round(float(groups.get_group('late').count()),2))
        res['SMIMERolloverLateMean'] = "{:,}".format(round(float(groups.get_group('late').mean()),2))
        res['SMIMERolloverLateMedian'] = "{:,}".format(round(float(groups.get_group('late').median()),2))
        res['SMIMERolloverLateStd'] = "{:,}".format(round(float(groups.get_group('late').std()),2))
    else:
        res['SMIMERolloverLateCount'] = "{:,}".format(0)
        res['SMIMERolloverLateMean'] = "{:,}".format(0)
        res['SMIMERolloverLateMedian'] = "{:,}".format(0)
        res['SMIMERolloverLateStd'] = "{:,}".format(0)
        
    if 'safe' in groups.indices:
        res['SMIMERolloverSafeCount'] = "{:,}".format(round(float(groups.get_group('safe').count()),2))
        res['SMIMERolloverSafeMean'] = "{:,}".format(round(float(groups.get_group('safe').mean()),2))
        res['SMIMERolloverSafeMedian'] = "{:,}".format(round(float(groups.get_group('safe').median()),2))
        res['SMIMERolloverSafeStd'] = "{:,}".format(round(float(groups.get_group('safe').std()),2))
    else:
        res['SMIMERolloverSafeCount'] = "{:,}".format(0)
        res['SMIMERolloverSafeMean'] = "{:,}".format(0)
        res['SMIMERolloverSafeMedian'] = "{:,}".format(0)
        res['SMIMERolloverSafeStd'] = "{:,}".format(0)

    df2 = df.loc[(df['type'].str.contains('late') | df['type'].str.contains('safe')) & (df['gap'] >= -52) & (df['gap'] <= 52)]
    df2_with_outlier = df.loc[(df['type'].str.contains('late') | df['type'].str.contains('safe'))]
    
    
    dfn = df2.loc[(df2['from user issuer'].str.contains('DFN') | df2['from user issuer'].str.contains('LUH')) & (df2['to user issuer'].str.contains('DFN') | df2['to user issuer'].str.contains('LUH'))]
    ca_change = df2.loc[(df2['from user issuer'] != df2['to user issuer'])]
   
   
    df2 = df2[['type', 'gap']]
    df2_with_outlier = df2_with_outlier[['type', 'gap']]
    df2_with_outlier = df2_with_outlier._to_pandas()
    
    if 'late' in groups.indices:
        late = groups.get_group('late')
        late = late.groupby('gap').size().reset_index(name='count')
        load_data.generate_plot_line(late['gap'], late['count'], "Gap between two certificates (Too Late)", "gap_between_certificates_late")
    if 'safe' in groups.indices:
        safe = groups.get_group('safe')
        safe = safe.groupby('gap').size().reset_index(name='count')
        load_data.generate_plot_line(safe['gap'], safe['count'], "Gap between two certificates (Safe)", "gap_between_certificates_safe")
    

    plt.clf()
    plt.figure(figsize=(6, 3))
    plt.grid(False)
    sns.set(rc={'figure.figsize': (6, 3)})
    sns.set_theme()
    sns.set_style("white")
    sns.set_style("ticks")
    colors = sns.color_palette("pastel")
    ax = df2['gap'].hist(bins=20, color=colors[2])
    ax.grid(False)

    plt.xlabel('weeks')
    plt.ylabel("# Certificates")
    sns.despine()
    plt.savefig(load_data.get_cache_path("graphs/paper/hist_key_rollover.png", testrun), bbox_inches='tight')

    plt.clf()
    plt.figure(figsize=(6, 3))
    plt.grid(False)
    sns.set(rc={'figure.figsize': (6, 3)})
    sns.set_theme()
    sns.set_style("white")
    sns.set_style("ticks")
    colors = sns.color_palette("pastel")
    maxVal = 23
    minVal = -55
    df2_with_outlier['gap'].clip(lower = minVal, upper= maxVal, inplace=True)
    ax = df2_with_outlier['gap'].hist(bins=24, color=colors[2])
    ax.grid(False)

    plt.xlabel('weeks')
    plt.xticks([-52,-40, -20, 0, 20])
    ax.set_xticklabels(["<=-52", "-40", "-20", "0", ">=20"])
    plt.ylabel("# Certificates")
    sns.despine()
    plt.savefig(load_data.get_cache_path("graphs/paper/hist_key_rollover_with_outlier.png", testrun), bbox_inches='tight')
    print(df2['gap'])
    print(df2_with_outlier['gap'])
    load_data.write_commands(res, load_data.get_cache_path('keyrollover.tex', testrun))
    return

    
def main():
    debug = True
    global testrun
    run_all = True

    extract_data = False


    overall = input("Run overall data? y/n\n")
    if overall == 'y':
        testrun = False
        print("> load all data...")

    print("> start analysis...")
    
    if run_all:
        print("> analysis multiple keys")
        analysis_multiple_keys()
        print("> analysis key roll overs")
        analysis_rollover(testrun)
        print("> analysis smime communication")
        analysis_smime_communication()
        print("> analysis certificates of university members")
        analysis_certificates(fileName='university_members')
        df = smime_user_all()
        print("> analysis all certificates")
        analysis_certificates(df, 'all')
        print("> generate table")
        smime_pgp_key_table()
    print("> done")
    

if __name__ == "__main__":
    main()
