from OpenSSL import *
from cryptography.hazmat.primitives.asymmetric import rsa, dsa, ec, ed25519, ed448
import cryptography
import re
import json

# RegEx needed for OID Parsing
oid_regex = re.compile(r"oid=(.*), name")


class Cert(object):
    """
    Convenient container object for human-readable and output-file friendly certificate contents
    """
    #pem = ''
    serial_number = '' #Needs pseudomization
    not_valid_before = ''
    not_valid_after = ''
    issuer_info = {}
    signature_hash_algorithm = ''
    key_size = ''
    key_type = ''
    crl_endpoints = -1
    ocsp_endpoint = False
    key_usage = ''
    extended_key_usage = []
    subject_alt_names = -1 #Should be E-Mail Count
    subject_alt_names_unique = -1 #Lowercased, to fetch case differences.
    error = 0

    def __repr__(self):
        json_out = {}
        json_out["sn"] = self.serial_number
        json_out["not_valid_b"] = str(self.not_valid_before)
        json_out["not_valid_a"] = str(self.not_valid_after)
        json_out["algo"] = self.signature_hash_algorithm
        json_out["size"] = self.key_size
        json_out["type"] = self.key_type
        json_out["issuer"] = self.issuer_info
        json_out["crl"] = self.crl_endpoints
        json_out["ocsp"] = self.ocsp_endpoint
        json_out["usage"] = self.key_usage
        json_out["ext_usage"] = self.extended_key_usage
        json_out["subj"] = self.subject_alt_names
        json_out["subju"] = self.subject_alt_names_unique
        if self.error != 0:
            json_out["error"] = str(self.error)
        return json.dumps(json_out, indent=None, separators=(',', ':'))

class MetaObject(object):
    level = 0
    def __repr__(self):
        json_out = {}
        json_out["level"] = self.level
        return json.dumps(json_out, indent=None, separators=(',', ':'))

def get_certificates(p7_data):
    """
    :param p7_data: OpenSSL.crypto.PKCS7
    :return: list of cryptography.x509.Certificate
    Returns all certificates for the PKCS7 structure, if present. Only
    objects of type ``signedData`` or ``signedAndEnvelopedData`` can embed
    certificates.
    See:
      https://stackoverflow.com/questions/16899247/how-can-i-decode-a-ssl-certificate-using-python
      https://www.programcreek.com/python/example/64094/OpenSSL.crypto.dump_certificate
      https://github.com/pyca/pyopenssl/pull/367/files#r67300900
    """
    certs = crypto._ffi.NULL
    if p7_data.type_is_signed():
        certs = p7_data._pkcs7.d.sign.cert
    elif p7_data.type_is_signedAndEnveloped():
        certs = p7_data._pkcs7.d.signed_and_enveloped.cert

    pycerts = []
    for i in range(crypto._lib.sk_X509_num(certs)):
        pycert = crypto.X509.__new__(crypto.X509)
        pycert._x509 = crypto._lib.sk_X509_value(certs, i)
        pycerts.append(crypto.X509.to_cryptography(pycert))

    if not pycerts:
        return None
    return pycerts

def dump(obj):
  for attr in dir(obj):
    print("obj.%s = %r" % (attr, getattr(obj, attr)))

def extract_smime_signature(payload):
    """
    Extract public certificates from the PKCS7 binary payload

    :param payload: bytes
    :return: list of Cert objects
    """
    pkcs7 = crypto.load_pkcs7_data(crypto.FILETYPE_ASN1, payload)
    certs = get_certificates(pkcs7)
    certList = []
    # Collect the following info from the certificates
    all_cert_times_valid = True
    for c in certs:
        # Convert to the modern & easier to use https://cryptography.io library objects
        c2 = crypto.X509.to_cryptography(c)
        c3 = Cert()

        c3.serial_number = c2.serial_number

        # check each certificate's time validity, ANDing cumulatively across each one
        c3.not_valid_before = c2.not_valid_before
        c3.not_valid_after = c2.not_valid_after


        try:
            for extension in c2.extensions:
                if extension.oid._name == "cRLDistributionPoints":
                    c3.crl_endpoints = len(extension.value._distribution_points) # Get counts for CRL Endpoints
                elif extension.oid._name == "authorityInfoAccess": #OCSP
                    for auth_info in extension.value:
                        if auth_info.access_method._name == "OCSP":
                            c3.ocsp_endpoint = True
                elif extension.oid._name == "keyUsage":
                    c3.key_usage = {}
                    c3.key_usage["content_commitment"] = extension.value.content_commitment
                    c3.key_usage["crl_sign"] = extension.value.crl_sign
                    c3.key_usage["data_encipherment"] = extension.value.data_encipherment
                    c3.key_usage["digital_signature"] = extension.value.digital_signature
                    c3.key_usage["decipher_only"] = extension.value._decipher_only
                    c3.key_usage["encipher_only"] = extension.value._encipher_only
                    c3.key_usage["key_agreement"] = extension.value.key_agreement
                    c3.key_usage["key_cert_sign"] = extension.value.key_cert_sign
                    c3.key_usage["key_encipherment"] = extension.value.key_encipherment
                elif extension.oid._name == "extendedKeyUsage":
                    e_key_usages=[]
                    for e_key_usage in extension.value:
                        e_key_usages.append(e_key_usage._name)
                    c3.extended_key_usage = e_key_usages;
                elif extension.oid._name == "subjectAltName":
                    list_of_names = set()
                    list_of_lowercase_names = set()
                    for names in extension.value:
                        list_of_names.add(names.value)
                        try:
                            list_of_lowercase_names.add(names.value.lower())
                        except: #Catch error, incase that we get a name object returned and write the Name Object into the set
                            list_of_lowercase_names.add(names.value)
                    #Grab names to identify duplicate entries, store numbers only
                    c3.subject_alt_names = len(list_of_names)
                    c3.subject_alt_names_unique = len(list_of_lowercase_names)
        except ValueError as e:
            c3.error = e #Store error message
            print(e) # Catch exceptions here.

        c_issuer_info = {}
        for i in c.issuer:
            c_issuer_info.update({i.oid._name: i.value})

        try:
            c3.signature_hash_algorithm = c2.signature_hash_algorithm.name
        except cryptography.exceptions.UnsupportedAlgorithm as e:
            #Error thrown looks like this: #"Signature algorithm OID:<ObjectIdentifier(oid=1.2.840.113549.1.1.2, name=Unknown OID)> not recognized"
            #Regex: oid = (.*), name

            encountered_oid = oid_regex.search(str(e)).group(1)

            known_oids={"1.2.840.113549.1.1.2": "md2WithRSAEncryption"}

            if encountered_oid in known_oids:
                c3.signature_hash_algorithm = known_oids[encountered_oid]
            else:
                c3.signature_hash_algorithm = encountered_oid #Write down oid, since we couldn't identify it on the fly
        pubkey = c.public_key()
        c3.key_size = pubkey.key_size
        if(isinstance(pubkey, rsa.RSAPublicKey)):
            c3.key_type = "RSA"
        elif (isinstance(pubkey, dsa.DSAPublicKey)):
            c3.key_type = "DSA"
        elif (isinstance(pubkey, ec.EllipticCurvePublicKey)):
            c3.key_type = "EC"
        elif (isinstance(pubkey, ed25519.Ed25519PublicKey)):
            c3.key_type = "ED25519"
        elif (isinstance(pubkey, ed448.Ed448PublicKey)):
            c3.key_type = "ED448"
        else:
            c3.key_type = "other"
        c3.issuer_info = c_issuer_info

        #The following line would get the PEM Format of the cert
        certList.append(c3)
    return certList