import gnupg
import random
import os
from parser import KeyWords
import base64

pgp_key_block = "-----BEGIN PGP PUBLIC KEY BLOCK-----"
pgp_signature_block = "-----BEGIN PGP SIGNATURE-----"
pgp_encryption_block = "-----BEGIN PGP ENCRYPTION-----"
pgp_message_block = "-----BEGIN PGP MESSAGE-----"


"""
About thread safety: https://www.gnupg.org/documentation/manuals/gpgme/Multi-Threading.html

We need keys = gpg.scan_keys(key_file_name)


dict key 	    dict value (all string values)
type 	        Type of key                                                         -> Store
trust 	        The validity of the key                                             -> ignore
length 	        The length of the key in bits                                       -> Store
algo 	        Public key algorithm                                                -> Store
keyid 	        The key ID                                                          -> Store (Pseudonym)
date 	        The creation date of the key in UTC as a Unix timestamp             -> Store (Category)
expires 	    The expiry date of the key in UTC as a timestamp, if specified      -> Store (Category)
dummy 	        Certificate serial number, UID hash or trust signature info         -> ignore
ownertrust 	    The level of owner trust for the key                                -> ignore
uid 	        The user ID                                                         -> ???? length?
sig 	        Signature class                                                     -> Store
cap 	        Key capabilities                                                    -> Store
issuer 	        Issuer information                                                  -> ignore ???
flag 	        A flag field                                                        -> ignore
token 	        Token serial number                                                 -> ignore
hash 	        Hash algorithm                                                      -> Store
curve 	        Curve name for elliptic curve cryptography (ECC) keys               -> Store
compliance 	    Compliance flags                                                    -> Store
updated 	    Last updated timestamp                                              -> Store
origin 	        Origin of keys                                                      -> ???
subkeys 	    A list containing [keyid, type] elements for each subkey            -> ignore
subkey_info 	A dictionary of subkey information keyed on subkey id               -> Store (see all above)

 See: https://git.gnupg.org/cgi-bin/gitweb.cgi?p=gnupg.git;a=blob_plain;f=doc/DETAILS
*** Field 6 - Creation date

    The creation date of the key is given in UTC.  For UID and UAT
    records, this is used for the self-signature date.  Note that the
    date is usually printed in seconds since epoch, however, we are
    migrating to an ISO 8601 format (e.g. "19660205T091500").  This is
    currently only relevant for X.509.  A simple way to detect the new
    format is to scan for the 'T'.  Note that old versions of gpg
    without using the =--fixed-list-mode= option used a "yyyy-mm-tt"
    format.


"""


def read_key(key, binaryData = False):
    """
    :param key: key
    :return: a list of parsed keys
    """
    if len(key) == 0:
        print("Empty key")
        return []
    # 1. Write data in file because gnupg reads the keys from a file...
    (f,name) = create_file(binaryData)
    f.write(key)
    f.close()
    # 2. extract key info
    gpg = gnupg.GPG()
    keys = gpg.scan_keys(name)
    # 3. Remove file
    os.remove(name)
    return keys

def create_file(binaryData):
    name = str(int(random.random() * (10 ** 10)))
    try:
        if binaryData:
            name = name + '.key'
            f = open(name , 'xb')
        else:
            name = name + '.asc'
            f = open(name, 'x')
        return [f, name]
    except:
        create_file(binaryData)

def verify_signature(signature):
    gpg = gnupg.GPG(options=['-vv'])
    verified = gpg.verify(signature)

    # Since we do not have the key and the payload, most information are stored in the error message.
    """
    Example:
       gpg: armor: BEGIN PGP SIGNATURE
   # off=0 ctb=89 tag=2 hlen=3 plen=563
   :signature packet: algo 1, keyid 2EF06E7444F3E96F
       version 4, created 1583767248, md5len 0, sigclass 0x01
       digest algo 10, begin of digest 50 ca
       hashed subpkt 33 len 21 (issuer fpr v4 814FF6BC50669E051133854E2EF06E7444F3E96F)
       hashed subpkt 2 len 4 (sig created 2020-03-09)
       subpkt 16 len 8 (issuer key ID 2EF06E7444F3E96F)
       data: [4096 bits]
   gpg: no signed data
   gpg: can't hash datafile: No data
       """
    result = verified.stderr
    result = result.split(",")
    fields = []
    for r in result:
        fields.extend(r.split('\n'))

    result = {}
    keywords = [': algo', 'keyid', 'version', 'sigclass', 'digest algo ','data']
    for f in fields:
        for keyword in keywords:
            if keyword in f:
                res = f.split(keyword)
                result[keyword] = res[-1].strip()

    return result

def decrypt(payload):
    gpg = gnupg.GPG(options=['-vv'])
    dec = gpg.decrypt(payload)
    result = dec.stderr
    result = result.split(",")
    fields = []
    for r in result:
        fields.extend(r.split('\n'))

    result = {}
    keywords = ['algo', 'keyid', 'version', 'sigclass', 'digest algo ', 'data']
    for f in fields:
        for keyword in keywords:
            if keyword in f:
                res = f.split(keyword)
                result[keyword] = res[-1].strip()

    return result

def parse_autocrypt_header(header):
    parts = header.split(';')
    if len(parts) == 0:
        return {}
    autocrypt = {}
    for part in parts:
        x = part.split("=",1)
        if len(x) != 2:
            return {}
        keyword = x[0].strip()
        if keyword == "prefer-encrypt": # TODO: More similar key words?
            autocrypt[KeyWords.ParserKeyWord.AUTOCRYPT_STATE] = x[1]
        elif keyword == "keydata":
            key_data = x[1]
            key_data = base64.b64decode("".join(key_data.split()))
            key = read_key(key_data, binaryData = True)
            autocrypt[KeyWords.ParserKeyWord.AUTOCRYPT_KEY] = key
        elif "addr" in keyword:
            pass
        else:
            print("[Autocrypt] Missing: ", keyword)
    return autocrypt

def demo():
    gpg = gnupg.GPG()
    file = "data/Alice(439EE43C)–Public.asc"
    keys = gpg.scan_keys(file)
    print(keys)
    return keys


def keydata1():
    return """mQINBFgNwGABEAC5KgyebtaoREglO01VZoKrvwl9JqpnmmlRNs9ViG3vY4aVIFz2UTvdBbZf
     y88tKInL8fU1Ph2wlnEOKZWji4w0+nQhbS7YdLFmPcpazfM8fBjqZdxlHLVemssDPvYMDvwm
     TrxUQNDqIXRNnA9krCmARtc2H8+ovRu1bT1TNv9DXQwex2NsyGlJ0c9mnvp9FGwJUr0/hqes
     XJ7dyhvd7Rn1iss2vKMbLHAZuI2jw9yyZGLhlhrQaTxDAayedOKuFr7vzAGsONSvxHviM1R0
     RSk0z/xwXoyjzkWc0Hy9hFl7iaaP0C96HfkBwcTh/oiKjLcsMn4555BorT2ErCy1VoKh6ZXS
     nz7aBo8++01sLo9v5Y3bFCNeQXp9FLfMKEbWC6V1J/lws2Rdg8mDA2J8pUaMDcIcUXfWfxCV
     6AFLCRVqytsxJhNBBs1+h/zNkwGE8ogEjdUJgsUN4cfjSuAC1iEmNJkp7EnLbmMpRizUHfxY
     xQ7oOsQKcF7BUKv9QloMl5f1rnCRkq+d0Qb+U0JXadti7Jrz7fAD3BVXDTT0JvRpU9pLbLqc
     MT/KkMxxfKw+P4UWI1phP1n94xdJ2CdjlTWOq7x2N0ZdpUteacjarZN5vp+k9QWg137BU1rZ
     Xvnh+CZa17Z15EsKVKdwkPWRqdCLtRoL2XylHgNSQ9X84tghEQARAQABtChPbGl2ZXIgV2ll
     c2UgPG9saXZlci53aWVzZUBmdS1iZXJsaW4uZGU+iQJXBBMBCgBBAhsDBQkHhh+ABQsJCAcD
     BRUKCQgLBRYCAwEAAh4BAheAFiEEgU/2vFBmngURM4VOLvBudETz6W8FAlhHxdwCGQEACgkQ
     LvBudETz6W/+fg/9HRsp1w7DMEyb+kPbczahqaZftykLzL/lndnt4BqPXF99FFdeBHoO08WY
     rol2gJmGpC5ji2W3Mnt6fHt6Bv4yDBiAWFNwSU28kvPOevDBPyVuRVqKpk7zwy3rDRWRqQW2
     /I2w1cbqnOxnMGn8ccYKiH+krrI4G72zN0DNrgQWGIRRfGFDOVLoSlHvi9qFgTstsv0EZsg9
     Q39S9u/OZFe+6W1NzQG4nrE5jZ88G5pMnjohMOpJNYJBjfC6ld+q3KXsEEqr/UvefoSv9lnA
     wjKK9cEtt3HLffBRIkMAWXLWbuX6crRLiGbmiUuZCHv+MfeinUGigmGhZzcLF6Z/FrJlGFRa
     WY4m4l1AwBq8b3MFTdMwnjhAg4Azl1g8Bx8607oxsHhRS+VLA4Qm6IxOI1bR82eBoWy+AAgi
     mviTDgGPmH/ByD+911GaL1utp52BajZeUrIRSRd2cCWCxv93fU802OVdWQLV5WRO08xfP4s0
     3PorC5WBnVqQ2QJd836P4A51wdim4BO7m3tGKxZyKPBxA4zvMA4cjbR4fbCo7LaM1muEKo4L
     e7J0UJEEUk6rYHcT83VmXNPInJwUZ0ywCSwaLmh8N8goqPfxi8eXl4qh6w9yo4mtW+5X++tb
     GdDHd5s+f1EjP1OjwiMNWP577Folp5cvEi8/cUfyLjYb4QbJMke5Ag0EWA3AYAEQAMwsnUSy
     BJ2kFulhNMGgWX+kk6vgTBpxXpm66R5vS8J3MP64yGM2VkELtGW1JK5gwF7NdAXp9thbb8ow
     IEPYW1r4iFKVERoq/YTySjfzjcOzr8nGMj9Cvg2h1RXbD8yIOxZmLoDwnVhrYPnB9xNuW8yb
     vEL2/Oq2d+1jPk9zmbjJR5qAGBhBbPN26izPa96p7An7K+NzTAWam+VwWPeRKGdtGIZbDCCp
     i1Zxd/toT7zcY0+J4RUd1w7BZqKD76YIGJF0hHco2wAoBc8/v0jYICu8YSlHH7rCrKsyie+8
     4Q4upfv/sVFLigY+ASqkGhCy3uYPv57mZuO3p2zzQ4GrkeoR5a0XgCe6UxQ7dy3D6BQgE+GU
     Ixp5ERJH4NN2D+6cKQBpbe8DsqE4qmWXQg8Isx36l/nnVaSnxMtQq3Z6tHPC1VfKu4l/y1gF
     zzXKxTKkYPb+tDdObGcLTIdda/Q4ANUWTKKgpPQz02B0TncbhYYUBj4BseIAFb6UGWG+qb2Q
     6oBPFAg1NFajXobJFu/RDoN3f+w3ry/SpbynCte2kRP+1oRM73c89JoWUbV+WJ1l6jjZQ6Ik
     7fIZtCzt/Qx+jHzSbS2ZVdkH489fBXOcPaLHWirWU40f4r8VAeB/4fDkcew1En6h4CF+GwhD
     S0ihC7tmWqbHBiVTYX5cEw2bqLx5ABEBAAGJAiUEGAEKAA8FAlgNwGACGwwFCQeGH4AACgkQ
     LvBudETz6W9T2A//ZoaABnRT/t7RZcC8lcmWRfem4yvXjLjlMaUeh/loVxmXu9r9hOg6Wm0e
     RWm1pImH+Q9E2veFBQHWGHycOFxBEo2Hj0ipeP5ZM7/HQbsTKtBPR5vPXpJq4w8gFlSdxCOy
     5ohL5QOF5JXVjHFV8DUBTokCun0V1IjwVRQT+qWHUxna1ABkKUiPIXfJNWBEj0LWPk+cndz/
     f3YEqtPvlGDf1TGxQDMe0XhbMFgP1sWgHKHwo1m5MKrczLbbbf4UMh/6fObijfwXLB9UoB1W
     lem19twLScM1ZtJ+NcEcIKFVMBGr86Qwsk6O9ioPcezzIsMhayobepx6A4N3t/s3I76i5oWn
     p075w9dpK8GaEHSSfITKj85IdXje9q+N3qQD5fWtZyvBXlMwFRpRgEgFSggHnaTJiF6CFxqb
     Zc5AEPULhpziToFoawdE7X08I2s0Swloe4BRxVTS9nAtdp9iaG3gMCRwiTrewZux5rvDyTRd
     /w2lbVOhRCVpP5HAHelARlShoQ9Viu6pABWnLLU2ulQtenoagv2l1EnJ583b+1EaajgWOFY2
     yLYxjaMwlBAcSfARLVmuHh4pKfOlb/Ir0tvNALiD6Rw8zquOqVKigpAaw0hS5eV6eSqdafeW
     O31IKCkg63j7QVjwLZlT9XiQjC6BsfBe+IP8DRNWT+cIdch2mU4="""

def keydata2():
    return """
    mDMEX+i03xYJKwYBBAHaRw8BAQdACA4xvL/xI5dHedcnkfViyq84doe8zFRid9jW7CC9XBiI0QQf
    FgoAgwWCX+i03wWJBZ+mAAMLCQcJEOCS6zpcoQ26RxQAAAAAAB4AIHNhbHRAbm90YXRpb25zLnNl
    cXVvaWEtcGdwLm9yZ/tr8E9NA10HvcAVlSxnox6z62KXCInWjZaiBIlgX6O5AxUKCAKbAQIeARYh
    BMKfigwB81402BaqXOCS6zpcoQ26AADZHQD/Zx9nc3N2kj13AUsKMr/7zekBtgfSIGB3hRCU74Su
    G44A/34Yp6IAkndewLxb1WdRSokycnaCVyrk0nb4imeAYyoPtBc8ZGtnQGZpZnRoaG9yc2VtYW4u
    bmV0PojRBBMWCgCDBYJf6LTfBYkFn6YAAwsJBwkQ4JLrOlyhDbpHFAAAAAAAHgAgc2FsdEBub3Rh
    dGlvbnMuc2VxdW9pYS1wZ3Aub3JnL0Gwxvypz2tu1IPG+yu1zPjkiZwpscsitwrVvzN3bbADFQoI
    ApsBAh4BFiEEwp+KDAHzXjTYFqpc4JLrOlyhDboAAPkXAP0Z29z7jW+YzLzPTQML4EQLMbkHOfU4
    +s+ki81Czt0WqgD/SJ8RyrqDCtEP8+E4ZSR01ysKqh+MUAsTaJlzZjehiQ24MwRf6LTfFgkrBgEE
    AdpHDwEBB0DkKHOW2kmqfAK461+acQ49gc2Z6VoXMChRqobGP0ubb4kBiAQYFgoBOgWCX+i03wWJ
    BZ+mAAkQ4JLrOlyhDbpHFAAAAAAAHgAgc2FsdEBub3RhdGlvbnMuc2VxdW9pYS1wZ3Aub3Jnfvo+
    nHoxDwaLaJD8XZuXiaqBNZtIGXIypF1udBBRoc0CmwICHgG+oAQZFgoAbwWCX+i03wkQPp1xc3He
    VlxHFAAAAAAAHgAgc2FsdEBub3RhdGlvbnMuc2VxdW9pYS1wZ3Aub3JnaheiqE7Pfi3Atb3GGTw+
    jFcBGOaobgzEJrhEuFpXREEWIQQttUkcnfDcj0MoY88+nXFzcd5WXAAAvrsBAIJ5sBg8Udocv25N
    stN/zWOiYpnjjvOjVMLH4fV3pWE1AP9T6hzHz7hRnAA8d01vqoxOlQ3O6cb/kFYAjqx3oMXSBhYh
    BMKfigwB81402BaqXOCS6zpcoQ26AADX7gD/b83VObe14xrNP8xcltRrBZF5OE1rQSPkMNy+eWpk
    eCwA/1hxiS8ZxL5/elNjXiWuHXEvUGnRoVj745Vl48sZPVYMuDgEX+i03xIKKwYBBAGXVQEFAQEH
    QIGex1WZbH6xhUBve5mblScGYU+Y8QJOomXH+rr5tMsMAwEICYjJBBgWCgB7BYJf6LTfBYkFn6YA
    CRDgkus6XKENukcUAAAAAAAeACBzYWx0QG5vdGF0aW9ucy5zZXF1b2lhLXBncC5vcmcEAx9vTD3b
    J0SXkhvcRcCr6uIDJwic3KFKxkH1m4QW0QKbDAIeARYhBMKfigwB81402BaqXOCS6zpcoQ26AAAX
    mwD8CWmukxwskU82RZLMk5fm1wCgMB5z8dA50KLw3rgsCykBAKg1w/Y7XpBS3SlXEegIg1K1e6dR
    fRxL7Z37WZXoH8AH
    """

def demoAutocrypt():
    data = "Autocrypt:  addr=xaskdljasoidjas@sdfsaf.com; prefer-encrypt=mutual; keydata="+keydata1()
    return parse_autocrypt_header(data)

