import hashlib, binascii
import re
from datetime import datetime

from dateutil.parser import parse

from parser import KeyWords
from parser import PGPHelper
from records.MailMapping import mailmapping, shortMap
from records.Types import Postbox
import json
import sys

import email.header

def decode_mime_words(s):
    #Used to handle Mailadresses starting with for example: =?utf-8?Q?
    return u''.join(
        word.decode(encoding or 'utf8') if isinstance(word, bytes) else word
        for word, encoding in email.header.decode_header(s))

def extract_address(value):
    """
    Extract only the email addresses according to https://tools.ietf.org/html/rfc5322#section-3.4
    Also change all to lower case
    :param value: may contain an email addresses
    :return: email addresses
    """
    regex = r"""\A(?:[a-z0-9!#$%&'*+/=?^_‘{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_‘{|}~-]+)*
 |  "(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]
      |  \\[\x01-\x09\x0b\x0c\x0e-\x7f])*")
@ (?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?
  |  \[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}
       (?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:
          (?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]
          |  \\[\x01-\x09\x0b\x0c\x0e-\x7f])+)
     \])\z"""
    addresses = re.findall(r'[\w\.-]+@[\w\.-]+', decode_mime_words(value))
    return list(set(a.lower() for a in addresses))


def hash(value, salt):
    value = str.encode(value)  # TODO: Are different encodings a probem?
    hashedValue = hashlib.pbkdf2_hmac('sha1', value, salt, 1)
    return binascii.hexlify(hashedValue).decode("utf-8")


def categorize_folder(folder):
    # Todo: check and extend key words
    inbox = ['inbox', 'postfach', 'eingang', 'posteingang']
    spam = ['junk', 'spam']
    sent = ['outgoing', 'sent', 'send', 'gesendet']
    trash = ['trash', 'papierkorb']
    trash_30d = ['30dtrash']

    name = folder.lower().split("/")[0]
    subfolder = folder.count("/")
    if name in inbox:
        if subfolder == 0:
            return Postbox.inb
        else:
            return Postbox.inb_sub
    elif name in spam:
        return Postbox.spam
    elif name in sent:
        if subfolder == 0:
            return Postbox.sent
        else:
            return Postbox.sent_sub
    elif name in trash:
        return Postbox.trash
    elif name in trash_30d:
        return Postbox.trash_30d
    else:
        return Postbox.Other


def categorize_date(date):
    """
    Converts a date input to the format Week:Year

    %W	Week number of the year (Monday as the first day of the week) as a decimal number. All days in a new year preceding the first Monday are considered to be in week 0.
    %Y	Year with century as a decimal number.

    Example:
    Input: Mon, 06 Jan 2020 20:00:16 +0100
    Output: 01:2020

    Input: Tue, 07 Jan 2020 09:00:17 +0100
    Output: 01:2020

    :param date:
    :return: week number : year
    """
    try:
        date_object = parse(date, fuzzy = True) #Fuzzy helps parsing more data without failing
        pseudo_date = date_object.strftime("%U:%Y")
        return pseudo_date
    except:
        return "1:1971"

def categorize_timestamp(ts):
    """
    Converts a timestamp input to the format Week:Year

    %W	Week number of the year (Monday as the first day of the week) as a decimal number. All days in a new year preceding the first Monday are considered to be in week 0.
    %Y	Year with century as a decimal number.

    Example:
    1595247041 -> 30:2020
    1610799041 -> 2:2021
    """

    ts = int(ts)
    date = datetime.fromtimestamp(ts)
    res = "" + str(date.isocalendar()[1]) + ":" + str(date.year)
    return res


def pseudo_body_structure(structure, salt):
    whitelist = ["signature", "smime"]
    critical_words = ["filename", "name"]
    for word in critical_words:
        regex = r'\(\"' + word + r'\" \"(\w*).\w*\"\)'
        names = re.findall(regex, structure)
        for f in names:
            if not (f in whitelist):
                pseudo = hash(f, salt)
                structure = structure.replace(f, pseudo)
    return structure


def pseudo_dkim(dkim, salt, priv_username="test", priv_domain="test"):
    """
    DKIM: (see: https://en.wikipedia.org/wiki/DomainKeys_Identified_Mail)
    v = version q = default query method
    a = signing algorithm c =  canonicalization algorithm(s) for header and body
    h = header fields - list of those that have been signed
    d = domain s = selector t = signature timestamp x = expire time
    skip:

    bh = body hash b = signature of headers and body
    z = copy of header information

    Example: dkim = 'v=1; a=rsa-sha256; c=relaxed/simple; d=example.net;'
    :param dkim:
    :return:
    """
    try:
        # Just to make parsing more easier...
        if not dkim.endswith(';'):
            dkim = dkim + ';'

        #Remove all whitespaces to make further regex easier and less error prone
        pseudo = re.sub(r"\s*", "", dkim)
        keep_keywords = ["a", "c", "d", "h", "i", "q", "t", "v", "x"]

        #Filter out potential privacy leaking keywords first
        skipped_keywords = ["s", "bh", "b", "z"]
        for key in skipped_keywords:
            pseudo = re.sub(key + r"=[^\s;]*;", "", pseudo)
        domain_pattern = r"d=([^\s;]*);"
        domains = re.findall(domain_pattern, pseudo)
        for domain in domains:
            hashed = hash(domain.strip(), salt)
            pseudo = re.sub(domain_pattern, "d=" + hashed + ";", pseudo)

        time_pattern = r"t=([^\s;]*);"
        times = re.findall(time_pattern, pseudo)
        for time in times:
            hashed = categorize_timestamp(time)
            pseudo = re.sub(time_pattern, "t=" + hashed + ";", pseudo)

        time2_pattern = r"x=([^\s;]*);"
        times = re.findall(time2_pattern, pseudo)
        for time in times:
            hashed = categorize_timestamp(time)
            pseudo = re.sub(time2_pattern, "x=" + hashed + ";", pseudo)

        i_pattern = r"i=([^\s;]*)@([^\s;]*);"
        s = re.findall(i_pattern, pseudo)
        for i in s:
            before_at = i[0].strip()
            if len(before_at) > 0:
                hashed = hash(before_at, salt)
            else:
                hashed = ""
            hashed_domain = hash(i[1].strip(), salt)
            pseudo = re.sub(i_pattern, "i=" + hashed + "@" + hashed_domain + ";", pseudo)

        #Only keep keywords we want
        pseudo_new = ""
        for key in keep_keywords:
            found_keyword = re.findall(key + r"=[^\s;]*;", pseudo)
            for keyword in found_keyword:
                if key != "i" and "@" in keyword:
                    print("[pseudo_dkim] Found @ in key where it shouldn't be. Skipping keyword. Keyword was:", key)
                    pseudo_new += key+"=skippedPseudoAt;"
                elif (priv_username in keyword or priv_domain in keyword):
                    print("[pseudo_dkim] Found username or domain in Key. Skipping keyword. Keyword was:", key)
                    pseudo_new += key+"=skippedPseudoPriv;"
                else:
                    pseudo_new += keyword


        '''
        expiry_pattern = r"x=\w*;"
        dates = re.findall(expiry_pattern, pseudo)
        for date in dates:
            # remove x= and ;
            ts = date[2:len(date) - 1]
            hashed = categorize_timestamp(ts)
            pseudo = re.sub(domain_pattern, "x=" + hashed + ";", pseudo)
        create_pattern = r"t=\w*;"
        dates = re.findall(create_pattern, pseudo)
        for date in dates:
            # remove x= and ;
            ts = date[2:len(date) - 1]
            hashed = categorize_timestamp(ts)
            pseudo = re.sub(domain_pattern, "t=" + hashed + ";", pseudo)
        '''
        return pseudo_new.strip()
    except Exception:
        return "dkimerror"


def pseudo_cert(cert, salt):
    """
    Removes possibly identifying information from the leaf certificate.
    SubjectName+SubjectAltName is not stored in the object, only a count.
    Issuer doesn't need to be pseudonymized, since they are public anyway.
    SerialNumber is hashed.
    Valid_before+Valid_after is changed to week of the year.
    """

    leaf_cert_number = len(cert)-1
    cert[leaf_cert_number].serial_number = hash(str(cert[leaf_cert_number].serial_number), salt)
    cert[leaf_cert_number].not_valid_before = categorize_date(str(cert[leaf_cert_number].not_valid_before))
    cert[leaf_cert_number].not_valid_after = categorize_date(str(cert[leaf_cert_number].not_valid_after))

    #If len == 1, then it's self signed. Remove issuer for privacy reasons.
    if len(cert) == 1:
        cert[leaf_cert_number].issuer_info = "selfsigned"

    pseudo = cert
    return pseudo


def pseudo_pgp(pgp_key, salt):
    """

    :param pgp_key: a pgp key we like to store
    :param salt: used salt
    :return: type of key, key length, public key/hash algo, curve name
    id (pseudo), creation/expiry/update date (category),
    # uids, signature class, key capabilities, compliance flags
    subkeys (recurive extraction)

    """
    # Lists of possible keywords. We like to track them.
    copy_keywords = ['type', 'length', 'algo', 'hash', 'sig', 'cap', 'curve', 'compliance']
    date_keywords = ['date', 'expires']
    pseudo_keywords = ['keyid']
    timestamp_keywords = ['updated']
    special_keywords = ['subkey_info', 'uid', 'uids']

    all_keywords = copy_keywords + date_keywords + pseudo_keywords + timestamp_keywords + special_keywords
    pseudo_key = {}

    for word in all_keywords:
        if word in pgp_key:
            value = pgp_key[word]
            # Avoid empty strings...
            if type(value) is str and value.strip() == '':
                continue

            if word in date_keywords:
                value = categorize_timestamp(value)
            elif word in timestamp_keywords:
                value = categorize_timestamp(value)
            elif word in pseudo_keywords:
                value = hash(value, salt)  # TODO: Value is very long.
            elif word in special_keywords:
                if word == "subkey_info":
                    new_value = {}
                    for subkey in value:
                        pseudo_subkey = hash(subkey, salt)
                        subkey_value = pseudo_pgp(value[subkey], salt)
                        new_value[pseudo_subkey] = subkey_value
                    value = new_value
                elif word == "uid" or word == "uids":
                    if type(value) is str:
                        value = 1
                    else:
                        value = len(value) # Trennung zwischen Groß und kleinschreibung. Siehe SMIME
                else:
                    print("[Pseudo PGP] Error. Not implemented: ", word) #value, word)
            pseudo_key[word] = value
    return pseudo_key


def pseudomize(mail, salt, username="", domain=""):
    """
    Pseudomize each elem of a mail.
    :param mail: a list of ParsedItems (keyword, value)
    :return: a list of pseudomized parsedItems.
    """
    # TODO Refactor?
    email_address_keywords = [KeyWords.JsonLevelKeyWord.USER, KeyWords.ParserKeyWord.USER, KeyWords.ParserKeyWord.TO,
                              KeyWords.ParserKeyWord.SENDER, KeyWords.ParserKeyWord.RETURN_PATH,
                              KeyWords.ParserKeyWord.DELIVERED_TO, KeyWords.ParserKeyWord.CC,
                              KeyWords.ParserKeyWord.BCC, KeyWords.ParserKeyWord.REPLY_TO]
    folder_keywords = [KeyWords.JsonLevelKeyWord.MAILBOX, KeyWords.ParserKeyWord.MAILBOX]
    date_keywords = [KeyWords.ParserKeyWord.HEADER_DATE, KeyWords.ParserKeyWord.RECEIVED_DATE]
    no_pseudo_keywords = [KeyWords.ParserKeyWord.USER_AGENT, KeyWords.ParserKeyWord.XMAILER]

    #Check for mailinglist headers
    mailinglist = False

    pseudo_mail = []
    counter_received_header = 0
    counter_virus_scans = 0

    all_pgp_keys = []

    for elem in mail:
        key = elem.key
        meta = elem.meta
        if key in email_address_keywords:
            try:
                addresses = extract_address(elem.value)
                for adr in addresses:
                    pseudo = hash(adr, salt)
                    domain = adr.rsplit("@")[1]
                    if adr.endswith("uni-hannover.de"):
                        if domain in mailmapping:
                            try:
                                pseudo = pseudo + "@" + shortMap[mailmapping[domain]] + ".h"
                            except:
                                pseudo = pseudo + "@err.h"
                        else:
                            #Search if this is a subdomain of a known domain and apply that mapping
                            subdomain_found = False
                            for d in mailmapping:
                                if domain.endswith("."+d) and d != "uni-hannover.de":
                                    try:
                                        pseudo = pseudo + "@" + shortMap[mailmapping[d]] + ".h"
                                    except:
                                        pseudo = pseudo + "@err.h"
                                    subdomain_found = True
                                    break
                            if not subdomain_found:
                                pseudo = pseudo + "@u.h" # Unknown domain
                                #print("[Pseudo] DEBUG-Domain: ", domain)
                    else:
                        pseudo = pseudo + "@ext"
                    pseudo_mail.append(KeyWords.ParsedItem(key, pseudo, meta))
            except Exception as e:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-mail"))
                print("[pseudo-method-mail] Error catched: ", e.__class__.__name__, e.__traceback__.tb_lineno, file=sys.stderr)
        elif key in folder_keywords:
            try:
                folder = categorize_folder(elem.value)
                pseudo_mail.append(KeyWords.ParsedItem(key, folder, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-folder"))
        elif key == KeyWords.ParserKeyWord.RECEIVED:
            counter_received_header += 1
        elif key == KeyWords.ParserKeyWord.Virus_SCAN:
            counter_virus_scans += 1
        elif key == KeyWords.ParserKeyWord.MSG_ID:
            try:
                msg_id = hash(elem.value, salt)
                pseudo_mail.append(KeyWords.ParsedItem(key, msg_id, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-msgid"))
        elif key in date_keywords:
            try:
                date = categorize_date(elem.value)
                pseudo_mail.append(KeyWords.ParsedItem(key, date, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-date"))
        elif key == KeyWords.JsonLevelKeyWord.BODY:
            pass
        elif key == KeyWords.ParserKeyWord.BODY_STRUCTURE or key == KeyWords.JsonLevelKeyWord.BODY_STRUCTURE:
            try:
                pseudo_structure = pseudo_body_structure(elem.value, salt)
                pseudo_mail.append(KeyWords.ParsedItem(key, pseudo_structure, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-body-structure"))
        elif key == KeyWords.ParserKeyWord.DKIM:
            # SKIP DKIM for this run, causes to much headache, not much insight
            #pseudo = pseudo_dkim(elem.value, salt, username, domain)
            #pseudo_mail.append(KeyWords.ParsedItem(key, pseudo, meta))
            pass
        elif key in no_pseudo_keywords:
            try:
                pseudo_mail.append(elem)
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-no-pseudo"))
        elif key == KeyWords.ParsedData.CERT_DATA:
            try:
                pseudo = pseudo_cert(elem.value, salt)
                pseudo_mail.append(KeyWords.ParsedItem(key, pseudo, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-cert-data"))
        elif key == KeyWords.ParsedData.SMIME_ENCRYPTION:
            # No personal data, only state of yes or no
            try:
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-smime-enc"))
        elif key == KeyWords.ParsedData.SMIME_SIGNATURE:
            # No personal data, only state of yes, no or unknown (encrypted)
            try:
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-smime-sig"))

        elif key == KeyWords.ParserKeyWord.AUTOCRYPT:
            # Autocrypt
            try:
                ac = PGPHelper.parse_autocrypt_header(elem.value)
                if len(ac) > 0:
                    state_word = KeyWords.ParserKeyWord.AUTOCRYPT_STATE
                    data_word = KeyWords.ParserKeyWord.AUTOCRYPT_KEY
                    ac_data = {}

                    if state_word in ac:
                        ac_data[str(state_word)] = ac[state_word]
                    if data_word in ac:
                        keys = ac[data_word]
                        pseudo_pgp_keys = []
                        for pgp_key in keys:
                            pseudo_pgp_keys.append(pseudo_pgp(pgp_key, salt))
                        ac_data[str(data_word)] = pseudo_pgp_keys
                        data = json.dumps(ac_data)
                        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParserKeyWord.AUTOCRYPT, data, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-autocrypt"))
        elif key == KeyWords.ParsedData.PGP_KEY_DATA_ATTACHMENT or key == KeyWords.ParsedData.PGP_KEY_DATA_INLINE:
            try:
                pseudo_pgp_key = []
                for pgp_key in elem.value:
                    pseudo_pgp_key.append(pseudo_pgp(pgp_key, salt))
                all_pgp_keys.append(pseudo_pgp_key)
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-pgp-key"))

        elif key == KeyWords.ParsedData.PGP_SIGNATURE_ATTACHMENT or key == KeyWords.ParsedData.PGP_SIGNATURE_INLINE:
            # No personal data except key id
            try:
                keyid = elem.value['keyid']
                keyid = hash(keyid, salt)
                elem.value['keyid'] = keyid
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
            except:
                elem.value['keyid'] = "-1"
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ERROR, "psigkeyid", meta))
        elif key == KeyWords.ParsedData.PGP_ENCRYPTION_ATTACHMENT or key == KeyWords.ParsedData.PGP_ENCRYPTION_INLINE:
            try:
                keyid = elem.value['keyid']
                keyid = hash(keyid, salt)
                elem.value['keyid'] = keyid
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
            except:
                elem.value['keyid'] = "-1"
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ERROR, "penckeyid", meta))
        elif key == KeyWords.ParsedData.PGP_INLINE_FORWARD:
            try:
                pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
            except:
                pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-pgp-inline-forward"))
        elif key == KeyWords.ParsedData.SMIME_ERROR:
            pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
        elif key == KeyWords.ParsedData.PGP_ERROR:
            pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
        elif key == KeyWords.ParsedData.OTHER_ERROR:
            pseudo_mail.append(KeyWords.ParsedItem(key, elem.value, meta))
        #Below is a list of currently unhandled keywords that would previously print out "Missing Matching"
        #We'll just skip those for now
        elif key == KeyWords.ParserKeyWord.FLAGS:
            pass
        elif key == KeyWords.ParserKeyWord.BODY:
            pass
        elif key == KeyWords.ParserKeyWord.AUTOCRYPT_KEY:
            print("[Pseudo] KeyWords.ParserKeyWord.AUTOCRYPT_KEY in unexpected location")
            pass
        elif key == KeyWords.ParserKeyWord.AUTOCRYPT_STATE:
            print("[Pseudo] KeyWords.ParserKeyWord.AUTOCRYPT_STATE in unexpected location")
            pass
        #One of the mailinglist headers
        elif key == KeyWords.ParserKeyWord.MAILINGLIST_LIST_HELP or key == KeyWords.ParserKeyWord.MAILINGLIST_LIST_SUBSCRIBE or key == KeyWords.ParserKeyWord.MAILINGLIST_LIST_UNSUBSCRIBE:
            mailinglist = True
        elif key == KeyWords.ParsedData.IS_MAILINGLIST:
            print("[Pseudo] KeyWords.ParsedData.IS_MAILINGLIST in unexpected location")
            pass
        else:
            # We're not handling this key -> dropped, no reason to throw an assert here.
            print("Missing matching")

    try:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.IS_MAILINGLIST, mailinglist))
    except:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-mailinglist"))

    try:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_KEY_DATA_ATTACHMENT, json.dumps(all_pgp_keys)))
    except:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-pgp-key-data-attachment-json"))

    try:
        received = KeyWords.ParsedItem(KeyWords.ParserKeyWord.RECEIVED, str(counter_received_header))
        pseudo_mail.append(received)
    except:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-received-count"))

    try:
        scan = KeyWords.ParsedItem(KeyWords.ParserKeyWord.Virus_SCAN, str(counter_virus_scans))
        pseudo_mail.append(scan)
    except:
        pseudo_mail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-method-virus-scan"))

    return pseudo_mail
