from typing import List
import json
import mailparser
import base64
import OpenSSL
import contextlib
import urllib.parse


from . import Pseudomizer, KeyWords, CertHelper, PGPHelper
from records.Types import SMIMEEnc, SMIMESig

DELIMITER = ":"
HEADER = "hdr."


# Possible flags: https://wiki2.dovecot.org/Tools/Doveadm/Fetch
# TODO: find detect inline pgp, flags
def is_multi_line_key_word(key):
    """
    checks if a key word matches to multiple lines
    :param key: a keyword from ParserKeyWord
    :return: boolean
    """
    multiline = [KeyWords.ParserKeyWord.RECEIVED.value]
    if key in multiline:
        return True
    else:
        return False


def add_hdr(key, is_json):
    """
    Some key words have a prefix (hdr.) in case of a text but not in case of a json file.
    :param key: a ParserKeyWord
    :param is_json: boolean
    :return: checks if hdr. has to be added to key.
    """
    hdr_elems = [KeyWords.ParserKeyWord.SENDER.value, KeyWords.ParserKeyWord.TO.value, KeyWords.ParserKeyWord.CC.value,
                 KeyWords.ParserKeyWord.BCC.value, KeyWords.ParserKeyWord.RECEIVED.value,
                 KeyWords.ParserKeyWord.MSG_ID.value, KeyWords.ParserKeyWord.DKIM.value,
                 KeyWords.ParserKeyWord.HEADER_DATE.value]
    result = key
    if is_json:
        return key.lower()
    if key in hdr_elems:
        result = HEADER + key
    return result.lower()


def parse_attachment(attachment, messageId = "default", parse_attachment = {}):
    level = -1
    if attachment["filename"] in parse_attachment:
        level = parse_attachment[attachment["filename"]]
    elif attachment["filename"]+".asc" in parse_attachment:
        level = parse_attachment[attachment["filename"]+".asc"]
    filename = attachment["filename"].lower()
    type = attachment["mail_content_type"].lower()

    if (filename.endswith(".asc")) or "pgp" in type:  # PGP
        payload = attachment["payload"]
        if attachment["content_transfer_encoding"].lower() == "base64":
            payload = base64.b64decode(payload)
        try:
            payload = str(payload)
            # Check for edgecase, where the result is actually a bytearray and parse it to be readable by GPG
            if payload.startswith("b'"):
                payload = payload.replace("b'", "").replace("'", "").replace("\\r", "\r").replace("\\n", "\n")
        except (UnicodeDecodeError, AttributeError):
            print("[parse_attachment] PGP Attachment unknown encoding, encountered error.")

        if PGPHelper.pgp_key_block in payload:
            try:
                key_data = PGPHelper.read_key(payload)
                return (KeyWords.ParsedData.PGP_KEY_DATA_ATTACHMENT, key_data, level),
            except:
                return (KeyWords.ParsedData.PGP_ERROR, "akey", level),
        elif PGPHelper.pgp_signature_block in payload:
            try:
                verified = PGPHelper.verify_signature(payload)
                return (KeyWords.ParsedData.PGP_SIGNATURE_ATTACHMENT, verified, level),
            except:
                return (KeyWords.ParsedData.PGP_ERROR, "asig", level),
        elif PGPHelper.pgp_encryption_block in payload or PGPHelper.pgp_message_block in payload:
            try:
                dec = PGPHelper.decrypt(payload)
                return (KeyWords.ParsedData.PGP_ENCRYPTION_ATTACHMENT, dec, level),
            except:
                return (KeyWords.ParsedData.PGP_ERROR, "adec", level),
        elif "PGP" in payload:
            print("Warning: Unknown PGP attachment.")
            return (KeyWords.ParsedData.PGP_ERROR, "aunk", level),
    elif filename.endswith(".p7s") or filename.endswith(".p7c"):  # SMIME
        payload = attachment["payload"]
        if attachment["content_transfer_encoding"].lower() == "base64":
            payload = base64.b64decode(payload)
        try:
            if filename.endswith(".p7s"):  # .p7s is used for signed mails
                return (KeyWords.ParsedData.CERT_DATA, CertHelper.extract_smime_signature(payload), level), (
                    KeyWords.ParsedData.SMIME_SIGNATURE, SMIMESig.sig, level)
            else:  # .p7c is used for certificate only
                return (KeyWords.ParsedData.CERT_DATA, CertHelper.extract_smime_signature(payload),level), (
                    KeyWords.ParsedData.SMIME_SIGNATURE, SMIMESig.un,level)
        except OpenSSL.crypto.Error as e:
            print("[OpenSSL] Cannot decode cert - possibly wrong encoding.")
            return (KeyWords.ParsedData.SMIME_ERROR, "adec", level),
    elif filename.endswith(".p7m"):  # SMIME Encrypted
        # Not much to do here, we can't extract the content, since it's encrypted
        return (KeyWords.ParsedData.SMIME_ENCRYPTION, SMIMEEnc.enc, level), (
            KeyWords.ParsedData.SMIME_SIGNATURE, SMIMESig.unknown, level)
    else:
        pass  # Ignore, no SMIME/PGP Stuff
    return None


def parse_line(line, isJson):
    split_line: List[str] = line.split(DELIMITER, 1)
    if len(split_line) == 2:
        key_word = split_line[0].strip().lower()

        value = split_line[1]
        for key in KeyWords.ParserKeyWord:
            if key_word == (add_hdr(key.value, isJson)):
                return (key, value)
    return (None, line)


def parse_mail(mail, isJson):
    """
    Main function expect a mail as string and outputs a list of parsed items (with a keyword and a value)
    :param mail: a string
    :return: a list of parsed items. Each has a keyword (key) and a value
    """

    autocrypt_finished = False
    dkim_finished = False
    to_finished = False
    cc_finished = False
    bcc_finished = False

    parsed_items = []
    for line in mail.splitlines():
        (key, value) = parse_line(line, isJson)

        if key == None and len(parsed_items) > 0:
            if is_multi_line_key_word(parsed_items[-1].key):
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value
            if parsed_items[-1].key == KeyWords.ParserKeyWord.AUTOCRYPT and not autocrypt_finished:
                if ":" in value:
                    # New unhandled tag found -> Autocrypt Header finished
                    autocrypt_finished = True
                    continue
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value
            if parsed_items[-1].key == KeyWords.ParserKeyWord.DKIM and not dkim_finished:
                if not(value.startswith("\t") or value.startswith(" ")):
                    # New unhandled tag found -> DKIM Header finished
                    dkim_finished = True
                    continue
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value

            if parsed_items[-1].key == KeyWords.ParserKeyWord.TO and not to_finished:
                if not (value.startswith("\t") or value.startswith(" ")):
                    # New unhandled tag found -> TO Header finished
                    to_finished = True
                    continue
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value
            if parsed_items[-1].key == KeyWords.ParserKeyWord.CC and not cc_finished:
                if not (value.startswith("\t") or value.startswith(" ")):
                    # New unhandled tag found -> CC Header finished
                    cc_finished = True
                    continue
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value
            if parsed_items[-1].key == KeyWords.ParserKeyWord.BCC and not bcc_finished:
                if not (value.startswith("\t") or value.startswith(" ")):
                    # New unhandled tag found -> BCC Header finished
                    bcc_finished = True
                    continue
                parsed_items[-1].value = parsed_items[-1].value + "\n" + value
        elif key != None:
            parsed_item = KeyWords.ParsedItem(key, value)
            parsed_items.append(parsed_item)

    return parsed_items

def parse_json_mail(mail):
    cryptoAttachments = {}
    try:
        structure = mail['imap.bodystructure']
        cryptoAttachments = parse_imap_structure(structure)
    except:
        cryptoAttachments = {}
    parsed_items = []

    # Temporary storage for header+body to throw into mailparser
    temp_header = ""
    temp_body = ""
    for elem in mail:
        value = mail[elem]
        if elem == KeyWords.JsonLevelKeyWord.USER.value:
            parsed_items.append(KeyWords.ParsedItem(KeyWords.JsonLevelKeyWord.USER, value))
        elif elem == KeyWords.JsonLevelKeyWord.MAILBOX.value:
            parsed_items.append(KeyWords.ParsedItem(KeyWords.JsonLevelKeyWord.MAILBOX, value))
        elif elem == KeyWords.JsonLevelKeyWord.HEADER.value:
            temp_header = value
            res = parse_mail(value, True)
            parsed_items.extend(res)
        elif elem == KeyWords.JsonLevelKeyWord.BODY.value:
            temp_body = value
        elif elem == KeyWords.JsonLevelKeyWord.BODY_STRUCTURE.value:
            parsed_items.append(KeyWords.ParsedItem(KeyWords.JsonLevelKeyWord.BODY_STRUCTURE, value))
        else:
            # TODO: Error handling!
            print("Warning: Encountered unknown keyword in parse_json_mail")


    # Mailparser writes a lot of warnings and errors into stderr/stdout. Redirect that output to /dev/null,
    # since we're only interested in the User-Agent and the Attachments. That makes the warnings unneccessary.
    with contextlib.redirect_stdout(None):
        with contextlib.redirect_stderr(None):
            # Parse mail with mailparser to get attachments
            mail = mailparser.parse_from_string(temp_header + temp_body)

    if mail.body != "":
        value = mail.body
        parsed_items.append(KeyWords.ParsedItem(KeyWords.JsonLevelKeyWord.BODY, value))
        for line in value.split("\n"):
            # Check if the PGP String starts at the beginning of a line. If not, then it's an answer/forwarded.
            if line.startswith(PGPHelper.pgp_key_block):
                try:
                    key_data = PGPHelper.read_key(value)
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_KEY_DATA_INLINE, key_data))
                except:
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ERROR, "ikey"))
            elif line.startswith(PGPHelper.pgp_signature_block):
                try:
                    verified = PGPHelper.verify_signature(value)
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_SIGNATURE_INLINE, verified))
                except:
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ERROR, "isig"))
            elif line.startswith(PGPHelper.pgp_encryption_block) or line.startswith(PGPHelper.pgp_message_block):
                try:
                    dec = PGPHelper.decrypt(value)
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ENCRYPTION_INLINE, dec))
                except:
                    parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_ERROR, "ienc"))
            elif (PGPHelper.pgp_key_block in line or PGPHelper.pgp_signature_block in line or
                  PGPHelper.pgp_encryption_block in line or PGPHelper.pgp_message_block in line):
                parsed_items.append(KeyWords.ParsedItem(KeyWords.ParsedData.PGP_INLINE_FORWARD, True))

    messageId = "unknown"
    if "Message-ID" in mail.headers:
        messageId = mail.headers["Message-ID"]

    mailagent_found = False
    #Thunderbird/most open source
    if "User-Agent" in mail.headers:
        parsed_items.append(KeyWords.ParsedItem(KeyWords.ParserKeyWord.USER_AGENT,
                                                mail.headers["User-Agent"].replace('\r', '').replace('\n', '')))
        mailagent_found = True
    #Outlook/Applemail
    if "X-Mailer" in mail.headers:
        parsed_items.append(KeyWords.ParsedItem(KeyWords.ParserKeyWord.XMAILER,
                                                mail.headers["X-Mailer"].replace('\r', '').replace('\n', '')))
        mailagent_found = True
    #Default GMail Client with external IMAP sets @email.android.com as message id. Doesn't work for gmail addresses
    if messageId.endswith("@email.android.com") and not mailagent_found:
        parsed_items.append(KeyWords.ParsedItem(KeyWords.ParserKeyWord.USER_AGENT, 'Android Mailclient via MessageID'))

    for attachment in mail.attachments:
        attachment_data = parse_attachment(attachment, messageId, cryptoAttachments)
        if attachment_data is not None:
            for key, value, meta in attachment_data:
                parsed_items.append(KeyWords.ParsedItem(key, value, meta))

    return parsed_items

def json_data_to_mails(data):
    mails = []
    for elem in data:
        try:
            res = parse_json_mail(elem)
            mails.append(res)
        except Exception as e:
            print("[json_data_to_mails] Exception catched: ", e.__class__.__name__)
            res = [KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "jdtm")]
            mails.append(res)
    return mails


def parse_json_string(string):
    string = string.replace('\'', '\"')
    data = json.loads(string)
    return json_data_to_mails(data)


def read_json_file(file):
    with open(file, 'r') as json_file:
        data = json.load(json_file)
        return data


def parse_json_file(file):
    data = read_json_file(file)
    return json_data_to_mails(data)


def parse_imap_structure(structure):
    counter = 0
    current_word = ""
    lookingForWord = False
    results = {}
    for c in structure:
        if c == "\"":
            lookingForWord = not lookingForWord
        elif lookingForWord:
            current_word = current_word + c
        else:
            if current_word.endswith(".asc") or current_word.endswith(".p7s") or current_word.endswith(".p7c") or current_word.endswith(".p7m"):
                #Remove URL encoding
                current_word = urllib.parse.unquote(current_word)
                #Remove starting utf-8''
                if current_word.startswith("utf-8''"):
                    current_word = current_word[7:]

                if current_word in results:
                    if results[current_word] >= counter:
                        results[current_word] = counter
                else:
                    results[current_word] = counter
            current_word = ""
            if c == "(":
                counter = counter + 1
            elif c == ")":
                counter = counter - 1
    return results